from hessQuik.networks import NN
from hessQuik.layers import resnetLayer
import hessQuik.activations as act
from copy import deepcopy


class resnetNN(NN):
    r"""
    Residual neural networks (ResNet) were popularized in the paper `Deep Residual Learning for Image Recognition`_ by He et al.
    Here, every layer is a single layer plus a skip connection.
    Let :math:`u_0` be the input into the ResNet.
    The construction is of the form

    .. _Deep Residual Learning for Image Recognition: https://ieeexplore.ieee.org/document/7780459

    .. math::

        \begin{align}
            u_1 &= u_0 + h\sigma(K_1 u_0 + b_1)\\
            u_2 &= u_1 + h\sigma(K_2 x + b_2)\\
                &\vdots \\
            u_{\ell} &= u_{\ell-1} + h\sigma(K_{\ell} u_{\ell-1} + b_{\ell})
        \end{align}

    where :math:`\ell` is the number of layers, called the depth of the network.
    Each vector of features :math:`u_i` is of size :math:`(n_s, w)` where :math:`n_s` is the number of samples
    and :math:`w` is the width of the network.
    Users choose the width and depth of the network and the activation function :math:`\sigma`.
    """

    def __init__(self, width: int, depth: int, h: float = 1.0,
                 act: act.hessQuikActivationFunction = act.softplusActivation(),
                 device=None, dtype=None):
        r"""

        :param width: dimension of hidden features
        :type width: int
        :param depth: number of ResNet layers
        :type depth: int
        :param h: step size, :math:`h > 0`. Default: 1.0
        :type h: float
        :param act: hessQuik activation function. Default: hessQuik.activations.softplusActivationFunction
        :type act: hessQuik.activations.hessQuikActivationFunction
        """
        factory_kwargs = {'device': device, 'dtype': dtype}

        args = ()
        for i in range(depth):
            args += (resnetLayer(width, h=h, act=deepcopy(act), **factory_kwargs),)

        super(resnetNN, self).__init__(*args)


if __name__ == '__main__':
    import torch
    from hessQuik.utils import input_derivative_check
    torch.set_default_dtype(torch.float64)

    # problem setup
    nex = 11
    d = 3
    x = torch.randn(nex, d)
    f = resnetNN(d, 4, h=0.5, act=act.softplusActivation())

    print('======= FORWARD =======')
    input_derivative_check(f, x, do_Hessian=True, verbose=True, forward_mode=True)

    print('======= BACKWARD =======')
    input_derivative_check(f, x, do_Hessian=True, verbose=True, forward_mode=False)



import unittest
import torch
import hessQuik.activations as act
import hessQuik.layers as lay
import hessQuik.networks as net
from utils import run_all_tests


class TestInputOutputDimensions(unittest.TestCase):

    @staticmethod
    def run_test(f, m, n):

        # check input and output dimensions
        assert f.dim_input() == m
        assert f.dim_output() == n

    def test_singleLayer(self):
        m = 3
        n = 4
        f = lay.singleLayer(m, n)

        print(self)
        self.run_test(f, m, n)

    def test_resnetLayer(self):
        m = 3
        d = 4
        f = lay.singleLayer(m, d)

        print(self)
        self.run_test(f, m, m)


if __name__ == '__main__':
    torch.set_default_dtype(torch.float64)
    unittest.main()



from .hessQuik_layer import hessQuikLayer
from .single_layer import singleLayer
from .resnet_layer import resnetLayer
from .icnn_layer import ICNNLayer
from .quadratic_layer import quadraticLayer
from .quadratic_icnn_layer import quadraticICNNLayer



import torch
import torch.nn as nn
import torch.nn.functional as F
import math
from hessQuik.layers import hessQuikLayer
import hessQuik.activations as act
from typing import Union, Tuple


class ICNNLayer(hessQuikLayer):
    r"""
    Evaluate and compute derivatives of a single layer.

    Examples::

        >>> import torch, hessQuik.layers as lay
        >>> f = lay.ICNNLayer(4, None, 7)
        >>> x = torch.randn(10, 4)
        >>> fx, dfdx, d2fd2x = f(x, do_gradient=True, do_Hessian=True)
        >>> print(fx.shape, dfdx.shape, d2fd2x.shape)
        torch.Size([10, 11]) torch.Size([10, 4, 11]) torch.Size([10, 4, 4, 11])

    """

    def __init__(self, input_dim: int, in_features: Union[int, None], out_features: int,
                 act: act.hessQuikActivationFunction = act.softplusActivation(),
                 device=None, dtype=None) -> None:
        r"""

        :param input_dim: dimension of network inputs
        :type input_dim: int
        :param in_features: number of input features. For first ICNN layer, set ``in_features = None``
        :type in_features: int or``None``
        :param out_features: number of output features
        :type out_features: int
        :param act: activation function
        :type act: hessQuikActivationFunction
        :var K: weight matrix for the network inputs of size :math:`(d, n_{out})`
        :var b: bias vector of size :math:`(n_{out},)`
        :var L: weight matrix for the input features of size :math:`(n_{in}, n_{out})`
        :var nonneg: pointwise function to force :math:`l` to have nonnegative weights. Default: ``torch.nn.functional.softplus``
        """
        factory_kwargs = {'device': device, 'dtype': dtype}
        super(ICNNLayer, self).__init__()

        self.input_dim = input_dim
        self.in_features = in_features
        self.out_features = out_features
        self.act = act

        # extract nonnegative weights
        self.nonneg = F.softplus

        self.K = nn.Parameter(torch.empty(input_dim, out_features, **factory_kwargs))

        if in_features is not None:
            self.L = nn.Parameter(torch.empty(in_features, out_features, **factory_kwargs))
        else:
            self.register_parameter('L', None)

        self.b = nn.Parameter(torch.empty(out_features, **factory_kwargs))
        self.reset_parameters()

    def reset_parameters(self):
        nn.init.kaiming_uniform_(self.K, a=math.sqrt(self.input_dim))

        if self.L is not None:
            nn.init.kaiming_uniform_(self.L, a=math.sqrt(self.in_features))
            fan_in, _ = nn.init._calculate_fan_in_and_fan_out(self.L)
        else:
            fan_in, _ = nn.init._calculate_fan_in_and_fan_out(self.K)

        bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
        nn.init.uniform_(self.b, -bound, bound)

    def dim_input(self) -> int:
        r"""
        number of input features + dimension of network inputs
        """
        n = self.input_dim
        if self.in_features is not None:
            n += self.in_features
        return n

    def dim_output(self) -> int:
        r"""
        number of output features + dimension of network inputs
        """
        return self.out_features + self.input_dim

    def forward(self, ux: torch.Tensor, do_gradient: bool = False, do_Hessian: bool = False, forward_mode: bool = True,
                dudx: Union[torch.Tensor, None] = None, d2ud2x: Union[torch.Tensor, None] = None) \
            -> Tuple[torch.Tensor, Union[torch.Tensor, None], Union[torch.Tensor, None]]:
        r"""
        Forward propagation through ICNN layer of the form

        .. math::

            f(x) =
            \left[\begin{array}{c} \sigma\left(\left[\begin{array}{c}u(x) & x\end{array}\right]
            \left[\begin{array}{c}L^+ \\ K\end{array}\right] + b\right) & x \end{array}\right]

        Here, :math:`u(x)` is the input into the layer of size :math:`(n_s, n_{in})` which is
        a function of the input of the network, :math:`x` of size :math:`(n_s, d)`.
        The output features, :math:`f(x)`, are of size :math:`(n_s, n_{out} + d)`.
        The notation :math:`(\cdot)^+` is a function that makes the weights of a matrix nonnegative.

        As an example, for one sample, :math:`n_s = 1`, the gradient with respect to
        :math:`\begin{bmatrix} u & x \end{bmatrix}` is of the form

        .. math::

            \nabla_x f = \text{diag}\left(\sigma'\left(\left[\begin{array}{c}u(x) & x\end{array}\right]
            \left[\begin{array}{c}L^+ \\ K\end{array}\right] + b\right)\right)
            \left[\begin{array}{c}(L^+)^\top & K^\top\end{array}\right]
            \left[\begin{array}{c}\nabla_x u \\ I\end{array}\right]

        where :math:`\text{diag}` transforms a vector into the entries of a diagonal matrix and :math:`I` is
        the :math:`d \times d` identity matrix.

        """

        (dfdx, d2fd2x) = (None, None)

        M = self.K
        if self.L is not None:
            M = torch.cat((self.nonneg(self.L), M), dim=0)

        z = ux @ M + self.b

        # forward pass
        f, dsig, d2sig = self.act.forward(z, do_gradient=do_gradient, do_Hessian=do_Hessian,
                                          forward_mode=True if forward_mode is True else None)
        f = torch.cat((f, ux[:, -self.input_dim:]), dim=1)

        if (do_gradient or do_Hessian) and forward_mode is True:
            dfdx = dsig.unsqueeze(1) * M

            # -------------------------------------------------------------------------------------------------------- #
            if do_Hessian:
                d2fd2x = (d2sig.unsqueeze(1) * M).unsqueeze(2) * M.unsqueeze(0).unsqueeze(0)

                # Gauss-Newton approximation
                if dudx is not None:
                    d2fd2x = dudx.unsqueeze(1) @ (d2fd2x.permute(0, 3, 1, 2) @ dudx.unsqueeze(1).permute(0, 1, 3, 2))
                    d2fd2x = d2fd2x.permute(0, 2, 3, 1)

                if d2ud2x is not None:
                    # extra term to compute full Hessian
                    d2fd2x += d2ud2x @ dfdx.unsqueeze(1)

                # concatenate zeros
                Z = torch.zeros(d2fd2x.shape[0], d2fd2x.shape[1], d2fd2x.shape[2], self.input_dim,
                                dtype=d2fd2x.dtype, device=d2fd2x.device)
                d2fd2x = torch.cat((d2fd2x, Z), dim=-1)
            # -------------------------------------------------------------------------------------------------------- #

            # finish computing gradient
            if dudx is not None:
                dfdx = dudx @ dfdx

            I = torch.ones(dfdx.shape[0], 1, 1, dtype=dfdx.dtype, device=dfdx.device) \
                * torch.eye(self.input_dim, dtype=dfdx.dtype, device=dfdx.device).unsqueeze(0)
            dfdx = torch.cat((dfdx, I), dim=-1)

        if (do_gradient or do_Hessian) and forward_mode is False:
            dfdx, d2fd2x = self.backward(do_Hessian=do_Hessian)

        return f, dfdx, d2fd2x

    def backward(self, do_Hessian: bool = False,
                 dgdf: Union[torch.Tensor, None] = None, d2gd2f: Union[torch.Tensor, None] = None)\
            -> Tuple[torch.Tensor, Union[torch.Tensor, None]]:
        r"""
        Backward propagation through ICNN layer of the form

        .. math::

            f(u) =
            \left[\begin{array}{c} \sigma\left(\left[\begin{array}{c}u & x\end{array}\right]
            \left[\begin{array}{c}L^+ \\ K\end{array}\right] + b\right) & x \end{array}\right]

        Here, the network is :math:`g` is a function of :math:`f(u)`.

        As an example, for one sample, :math:`n_s = 1`, the gradient of the network with respect to :math:`u` is of the form

        .. math::

            \nabla_{[u,x]} g = \left(\sigma'\left(\left[\begin{array}{c}u & x\end{array}\right]
            \left[\begin{array}{c}L^+ \\ K\end{array}\right] + b\right) \odot \nabla_{[f, x]} g\right)
            \left[\begin{array}{c}(L^+)^\top & K^\top\end{array}\right]

        where :math:`\odot` denotes the pointwise product.

        """
        M = self.K
        if self.L is not None:
            M = torch.cat((self.nonneg(self.L), M), dim=0)

        # obtain stored information from backward pass
        d2gd2ux = None
        dsig, d2sig = self.act.backward(do_Hessian=do_Hessian)

        # compute gradient
        dgdux = dsig.unsqueeze(1) * M

        # augment gradient
        M2 = torch.ones(dgdux.shape[0], 1, 1, dtype=dgdux.dtype, device=dgdux.device) \
            * torch.eye(self.input_dim, dtype=dgdux.dtype, device=dgdux.device).unsqueeze(0)

        if self.in_features is not None:
            Z = torch.zeros(dgdux.shape[0], self.input_dim, self.in_features)
            M2 = torch.cat((Z, M2), dim=-1).permute(0, 2, 1)

        dgdux = torch.cat((dgdux, M2), dim=-1)

        if do_Hessian:
            # TODO: change order of operations, multiply K's first; check if logic with better naming
            d2gd2ux = (d2sig.unsqueeze(1) * M.unsqueeze(0)).unsqueeze(2) * M.unsqueeze(0).unsqueeze(0)

            # concatenate zeros
            Z = torch.zeros(d2gd2ux.shape[0], d2gd2ux.shape[1], d2gd2ux.shape[2], self.input_dim,
                            dtype=d2gd2ux.dtype, device=d2gd2ux.device)
            d2gd2ux = torch.cat((d2gd2ux, Z), dim=-1)

            if d2gd2f is not None:
                # Gauss-Newton approximation
                h1 = (dgdux.unsqueeze(1) @ d2gd2f.permute(0, 3, 1, 2) @ dgdux.permute(0, 2, 1).unsqueeze(1))
                h1 = h1.permute(0, 2, 3, 1)

                # extra term to compute full Hessian
                N, _, _, m = d2gd2ux.shape
                h2 = d2gd2ux.view(N, -1, m) @ dgdf.view(N, m, -1)
                h2 = h2.view(h1.shape)

                # combine
                d2gd2ux = h1 + h2

        # finish computing gradient
        if dgdf is not None:
            dgdux = dgdux @ dgdf

        return dgdux, d2gd2ux


if __name__ == '__main__':
    from hessQuik.utils import input_derivative_check
    torch.set_default_dtype(torch.float64)

    nex = 11  # no. of examples
    d = 3  # no. of input features
    m = 5  # no. of output features
    x = torch.randn(nex, d)
    f = ICNNLayer(d, None, m, act=act.softplusActivation())

    print('======= FORWARD =======')
    input_derivative_check(f, x, do_Hessian=True, verbose=True, forward_mode=True)

    print('======= BACKWARD =======')
    input_derivative_check(f, x, do_Hessian=True, verbose=True, forward_mode=False)



import torch
import torch.nn as nn
import math
from hessQuik.layers import hessQuikLayer


class quadraticLayer(hessQuikLayer):
    r"""
    Evaluate and compute derivatives of a ICNN quadratic layer.

    Examples::

        >>> import hessQuik.layers as lay
        >>> f = lay.quadraticLayer(4, 2)
        >>> x = torch.randn(10, 4)
        >>> fx, dfdx, d2fd2x = f(x, do_gradient=True, do_Hessian=True)
        >>> print(fx.shape, dfdx.shape, d2fd2x.shape)
        torch.Size([10, 1]) torch.Size([10, 4, 1]) torch.Size([10, 4, 4, 1])

    """
    """
    f(x) = x @ v + 0.5 * x.t() @ A.t() @ A @ x + mu
    """

    def __init__(self, in_features: int, rank: int, device=None, dtype=None) -> None:
        r"""
        :param in_features: number of input features, :math:`n_{in}`
        :type in_features: int
        :param rank: number of columns of quadratic matrix, :math:`r`.  In practice, :math:`r < n_{in}`
        :type rank: int
        :var v: weight vector for network inputs of size :math:`(d,)`
        :var A: weight matrix for quadratic term of size :math:`(d, r)`
        :var mu: additive scalar bias
        """

        factory_kwargs = {'device': device, 'dtype': dtype}
        super(quadraticLayer, self).__init__()

        self.in_features = in_features
        self.rank = rank
        self.ctx = None

        # create final layer
        self.v = nn.Parameter(torch.empty(self.in_features, **factory_kwargs))
        self.mu = nn.Parameter(torch.empty(1, **factory_kwargs))
        self.A = nn.Parameter(torch.empty(self.rank, self.in_features, **factory_kwargs))
        self.reset_parameters()

    def reset_parameters(self):
        bound = 1 / math.sqrt(self.in_features)
        nn.init.uniform_(self.v, a=-bound, b=bound)
        nn.init.uniform_(self.mu)
        bound = 1 / math.sqrt(self.in_features)
        nn.init.uniform_(self.A, a=-bound, b=bound)

    def dim_input(self) -> int:
        """
        number of input features
        """
        return self.in_features

    def dim_output(self) -> int:
        """
        scalar
        """
        return 1

    def forward(self, u, do_gradient=False, do_Hessian=False, forward_mode=True, dudx=None, d2ud2x=None):
        r"""
        Forward propagation through quadratic layer of the form, for one sample :math:`n_s = 1`,

        .. math::

            f(x) = u(x) v + \frac{1}{2} u(x)  A A^\top  u(x)^\top + \mu

        Here, :math:`u(x)` is the input into the layer of size :math:`(n_s, n_{in})` which is
        a function of the input of the network, :math:`x`.
        The output features, :math:`f(x)`, are of size :math:`(n_s, 1)`.

        The gradient with respect to :math:`x` is of the form

        .. math::

                \nabla_x f = (v^\top + u A A^\top) \nabla_x u

        """
        (df, d2f) = (None, None)
        AtA = self.A.t() @ self.A
        f = u @ self.v + 0.5 * torch.sum((u @ AtA) * u, dim=1) + self.mu

        if (do_gradient or do_Hessian) and forward_mode is True:
            df = self.v.unsqueeze(0) + u @ AtA

            if do_Hessian:
                d2f = AtA

                if d2ud2x is not None:
                    d2f = dudx @ d2f @ dudx.permute(0, 2, 1)
                    d2f += (d2ud2x @ df.unsqueeze(1).unsqueeze(-1)).squeeze()

                d2f = d2f.unsqueeze(-1)

            # finish computing gradient
            df = df.unsqueeze(-1)

            if dudx is not None:
                df = dudx @ df

        if (do_gradient or do_Hessian) and forward_mode is not True:
            self.ctx = (u,)
            if forward_mode is False:
                df, d2f = self.backward(do_Hessian=do_Hessian)

        return f.unsqueeze(-1), df, d2f

    def backward(self, do_Hessian=False, dgdf=None, d2gd2f=None):
        r"""
        Backward propagation through quadratic ICNN layer of the form, for one sample :math:`n_s = 1`,

        .. math::

                f(u) = u v + \frac{1}{2} u A A^\top  u^\top + \mu

        Here, the network is :math:`g` is a function of :math:`f(u)`.

        The gradient of the layer with respect to :math:`u` is of the form

        .. math::

            \nabla_u f = v^\top + u A A^\top.

        """
        d2f = None

        x = self.ctx[0]
        AtA = self.A.t() @ self.A  # TODO: recompute this or store it?
        df = self.v.unsqueeze(0) + torch.matmul(x, AtA)

        if do_Hessian:
            # TODO: improve wasteful storage
            d2f = (torch.ones(x.shape[0], 1, 1, dtype=AtA.dtype, device=AtA.device) * AtA).unsqueeze(-1)

        return df.unsqueeze(-1), d2f


if __name__ == '__main__':
    from hessQuik.utils import input_derivative_check
    torch.set_default_dtype(torch.float64)

    # problem setup
    nex = 11  # no. of examples
    d = 4  # no. of input dimensiona features
    m = 7  # rank
    x = torch.randn(nex, d)
    f = quadraticLayer(d, m)

    print('======= FORWARD =======')
    input_derivative_check(f, x, do_Hessian=True, verbose=True, forward_mode=True)

    print('======= BACKWARD =======')
    input_derivative_check(f, x, do_Hessian=True, verbose=True, forward_mode=False)






from hessQuik.networks import NN
from hessQuik.layers import ICNNLayer
import hessQuik.activations as act
from typing import Union, Tuple, List
from copy import deepcopy


class ICNN(NN):
    r"""
    Input Convex Neural Networks (ICNN) were proposed in the paper `Input Convex Neural Networks`_ by Amos, Xu, and Kolter.
    The network is constructed such that it is convex with respect to the network inputs :math:`x`.
    It is constructed via

    .. _Input Convex Neural Networks: https://arxiv.org/abs/1609.07152

    .. math::

        \begin{align}
            u_1 &= \sigma(K_1 x + b_1)\\
            u_2 &= \sigma(L_2^+ u_1 + K_2 x + b_2)\\
                &\vdots\\
            u_{\ell} &= \sigma(L_{\ell}^+ u_{\ell-1} + K_{\ell} x + b_{\ell})
        \end{align}

    where :math:`\ell` is the number of layers.
    Here, :math:`(\cdot)^+` is a function that forces the matrix to have only nonnegative entries.
    The activation function :math:`\sigma` must be convex and non-decreasing.
    """

    def __init__(self, input_dim: int, widths: Union[Tuple, List],
                 act: act.hessQuikActivationFunction = act.softplusActivation(),
                 device=None, dtype=None):
        r"""

        :param input_dim: dimension of input features
        :type input_dim: int
        :param widths: dimension of hidden features
        :type widths: tuple or list
        :param act: hessQuik activation function. Default: hessQuik.activations.softplusActivationFunction
        :type act: hessQuik.activations.hessQuikActivationFunction
        """
        factory_kwargs = {'device': device, 'dtype': dtype}

        args = ()
        for i, w in enumerate(range(1, len(widths))):
            args += (ICNNLayer(input_dim, widths[i], widths[i + 1], act=deepcopy(act), **factory_kwargs),)

        super(ICNN, self).__init__(*args)


if __name__ == '__main__':
    import torch
    from hessQuik.utils import input_derivative_check
    torch.set_default_dtype(torch.float64)

    # problem setup
    nex = 11  # no. of examples
    d = 3  # no. of input features
    ms = [None, 5, 2, 7]  # no. of output features
    x = torch.randn(nex, d)

    f = ICNN(d, ms, act=act.quadraticActivation())

    print('======= FORWARD =======')
    input_derivative_check(f, x, do_Hessian=True, verbose=True, forward_mode=True)

    print('======= BACKWARD =======')
    input_derivative_check(f, x, do_Hessian=True, verbose=True, forward_mode=False)




from .hessQuik_activation_function import hessQuikActivationFunction
from .antiTanh_activation import antiTanhActivation
from .identity_activation import identityActivation
from .quadratic_activation import quadraticActivation
from .softplus_activation import softplusActivation
from .tanh_activation import tanhActivation
from .sigmoid_activation import sigmoidActivation



from line_profiler import LineProfiler
import random
import torch
import hessQuik.activations as act
import hessQuik.layers as lay
import hessQuik.networks as net

# f = net.fullyConnectedNN([10, 20, 1], act=act.antiTanhActivation())

width = 20
f = net.NN(lay.singleLayer(2, width, act=act.antiTanhActivation()),
           net.resnetNN(width=width, depth=4, act=act.antiTanhActivation()),
           lay.singleLayer(width, 1, act=act.identityActivation()))

x = torch.randn(100, f.dim_input())

f = net.NNPytorchAD(f)
x.requires_grad = True

def eval(x):
    out = f(x, do_gradient=True, do_Hessian=True, forward_mode=False)


# numbers = [random.randint(1, 100) for i in range(1000)]
lp = LineProfiler()
lp_wrapper = lp(eval)
lp.add_function(f.forward)
# lp.add_function(f[1][0].forward)
# lp.add_function(f[1][0].backward)
lp_wrapper(x)
lp.print_stats()




import torch
import cProfile
import profile
import hessQuik.activations as act
import hessQuik.layers as lay
import hessQuik.networks as net
from line_profiler import LineProfiler

nex = 100
d = 1
width = 100

# f = net.NN(lay.singleLayer(d, width, act=act.tanhActivation()),
#            lay.singleLayer(width, 32, act=act.tanhActivation()))
f = net.fullyConnectedNN([d, width, 1], act=act.antiTanhActivation())
x = torch.randn(nex, f.dim_input())


def eval():
    f0, df0, d2f0 = f(x, do_gradient=True, do_Hessian=True, forward_mode=True)


lp = LineProfiler()
lp.timer_unit = 1e-6
lp.add_function(eval)
# lp.add_function(f[0].forward)
lp.print_stats()

# nex = 100
# in_features = 4
# width = 100
# depth = 4
# out_features = 1
# f = net.NN(lay.singleLayer(in_features, width, act=act.identityActivation()),
#            net.resnetNN(width, depth, h=0.5, act=act.identityActivation()),
#            lay.singleLayer(width, out_features, act=act.identityActivation()))
#
# x = torch.randn(nex, f.dim_input())

# f = net.NNPytorchAD(f)
# x.requires_grad = True

# print('------ cProfile FORWARD ------')
# cProfile.run('f(x, do_gradient=True, do_Hessian=True)', sort='tottime')

# print('------ profile FORWARD------')
# profile.run('f(x, do_gradient=True, do_Hessian=True)', sort='tottime')

# print('------ torch FORWARAD ------')
# with torch.profiler.profile() as prof:
#     with torch.profiler.record_function("singleLayer"):
#         f(x, do_gradient=True, do_Hessian=True)
#
# print(prof.key_averages().table(sort_by="self_cpu_time_total", row_limit=-1))

# print('------ torch  ------')
# with torch.profiler.profile() as prof:
#     with torch.profiler.record_function("singleLayer"):
#         f0, df0, d2f0 = f(x, do_gradient=True, do_Hessian=True)
#
# print(prof.key_averages().table(sort_by="self_cpu_time_total", row_limit=-1))


# if __name__ == '__main__':
#     eval()



from .utils import convert_to_base, extract_data, insert_data
from .input_derivative_check import input_derivative_check, input_derivative_check_finite_difference
from .network_derivative_check import network_derivative_check
from .timing import timing_test
from .data import peaks
from .training import train_one_epoch, test, print_headers



import torch
from torch import Tensor
from typing import Tuple, Optional


def peaks(y: Tensor, do_gradient: bool = False, do_Hessian: bool = False) -> \
        Tuple[Tensor, Optional[Tensor], Optional[Tensor]]:
    r"""
    Generate data from the `MATLAB 2D peaks function`_

    .. _MATLAB 2D peaks function: https://www.mathworks.com/help/matlab/ref/peaks.html

    Examples::

         import matplotlib.pyplot as plt
         from matplotlib import cm
         from mpl_toolkits import mplot3d
         x, y = torch.linspace(-3, 3, 100), torch.linspace(-3, 3, 100)
         grid_x, grid_y = torch.meshgrid(x, y)
         grid_xy = torch.concat((grid_x.reshape(-1, 1), grid_y.reshape(-1, 1)), dim=1)
         grid_z, *_ = peaks(grid_xy)
         fig = plt.figure()
         ax = plt.axes(projection='3d')
         surf = ax.plot_surface(grid_x, grid_y, grid_z.reshape(grid_x.shape), cmap=cm.viridis)
         ax.set_xlabel('x')
         ax.set_ylabel('y')
         ax.set_zlabel('z')
         plt.show()

    :param y: (x, y) coordinates with shape :math:`(n_s, 2)` where :math:`n_s` is the number of samples
    :type y: torch.Tensor
    :param do_gradient: If set to ``True``, the gradient will be computed during the forward call. Default: ``False``
    :type do_gradient: bool, optional
    :param do_Hessian: If set to ``True``, the Hessian will be computed during the forward call. Default: ``False``
    :type do_Hessian: bool, optional
    :return:
            - **f** (*torch.Tensor*) - value of peaks function at each coordinate with shape :math:`(n_s, 1)`
            - **dfdx** (*torch.Tensor* or ``None``) - value of gradient at each coordinate  with shape :math:`(n_s, 2)`
            - **d2fd2x** (*torch.Tensor* or ``None``) - value of Hessian at each coordinate with shape :math:`(n_s, 4)`

    """

    df, d2f = None, None

    # function
    e1 = torch.exp(-1.0 * (y[:, 0]**2 + (y[:, 1] + 1)**2))
    f1 = 3 * (1 - y[:, 0])**2 * e1

    e2 = torch.exp(-y[:, 0]**2 - y[:, 1]**2)
    f2 = -10 * (y[:, 0] / 5 - y[:, 0]**3 - y[:, 1]**5) * e2

    e3 = torch.exp(-(y[:, 0] + 1)**2 - y[:, 1]**2)
    f3 = -(1 / 3) * e3
    f = f1 + f2 + f3

    if do_gradient or do_Hessian:
        # gradient
        dexp1x = -2 * y[:, 0]
        dexp1y = -2 * (y[:, 1] + 1)

        df1x = -6 * (1 - y[:, 0]) * e1 + dexp1x * f1
        df1y = dexp1y * f1

        dexp2x = -2 * y[:, 0]
        dexp2y = -2 * y[:, 1]
        df2x = -10 * (1 / 5 - 3 * y[:, 0]**2) * e2 + dexp2x * f2
        df2y = 50 * y[:, 1]**4 * e2 + dexp2y * f2

        dexp3x = -2 * (y[:, 0] + 1)
        dexp3y = -2 * y[:, 1]
        df3x = dexp3x * f3
        df3y = dexp3y * f3

        dfx = (df1x + df2x + df3x).unsqueeze(-1)
        dfy = (df1y + df2y + df3y).unsqueeze(- 1)

        df = torch.cat((dfx, dfy), dim=1)
        df = df.unsqueeze(-1)

        if do_Hessian:
            # Hessian
            dexp1xx = -2
            dexp1yy = -2
            d2f1xx = 6 * e1 + -6 * (1 - y[:, 0]) * e1 * dexp1x + dexp1xx * f1 + dexp1x * df1x
            d2f1xy = dexp1y * df1x
            d2f1yy = dexp1yy * f1 + dexp1y * df1y

            dexp2xx = -2
            dexp2yy = -2
            d2f2xx = 60 * y[:, 0] * e2 + -10 * (1 / 5 - 3 * y[:, 0] ** 2) * e2 * dexp2x + dexp2xx * f2 + dexp2x * df2x
            d2f2xy = 50 * y[:, 1] ** 4 * e2 * dexp2x + dexp2y * df2x
            d2f2yy = 200 * y[:, 1] ** 3 * e2 + 50 * y[:, 1] ** 4 * e2 * dexp2y + dexp2yy * f2 + dexp2y * df2y

            dexp3xx = -2
            dexp3yy = -2
            d2f3xx = dexp3x * df3x + dexp3xx * f3
            d2f3xy = dexp3x * df3y
            d2f3yy = dexp3y * df3y + dexp3yy * f3

            d2fxx = (d2f1xx + d2f2xx + d2f3xx).unsqueeze(-1)
            d2fxy = (d2f1xy + d2f2xy + d2f3xy).unsqueeze(-1)
            d2fyy = (d2f1yy + d2f2yy + d2f3yy).unsqueeze(-1)

            d2f = torch.cat((torch.cat((d2fxx, d2fxy), dim=1).unsqueeze(-1),
                             torch.cat((d2fxy, d2fyy), dim=1).unsqueeze(-1)), dim=-1)

            d2f = d2f.unsqueeze(-1)

    return f.unsqueeze(-1), df, d2f


if __name__ == "__main__":
    from hessQuik.tests import derivative_check
    torch.set_default_dtype(torch.float64)

    x = -3 + 6 * torch.rand(11, 2)

    f0, df0, d2f0 = peaks(x, do_gradient=True, do_Hessian=True)

    # directional derivative
    dx = torch.randn_like(x)
    dfx = torch.matmul(df0.transpose(1, 2), dx.unsqueeze(2)).squeeze(2)

    curvx = None
    if d2f0 is not None:
        curvx = torch.sum(dx.unsqueeze(2).unsqueeze(3) * d2f0 * dx.unsqueeze(1).unsqueeze(3), dim=(1, 2))

    derivative_check(peaks, x, dx, f0, dfx, curvx, verbose=True)



import torch
import torch.nn.functional as F
from hessQuik.activations import hessQuikActivationFunction


class softplusActivation(hessQuikActivationFunction):
    r"""
    Applies the softplus activation function to each entry of the incoming data.

    Examples::

        >>> import hessQuik.activations as act
        >>> act_func = act.softplusActivation()
        >>> x = torch.randn(10, 4)
        >>> sigma, dsigma, d2sigma = act_func(x, do_gradient=True, do_Hessian=True)

    """

    def __init__(self, beta: float = 1.0, threshold: float = 20.0) -> None:
        r"""

        :param beta: parameter affecting steepness of the softplus function.  Default: 1.0
        :type beta: float
        :param threshold: parameter for numerical stability.  Uses identity function when :math:`\beta x > threshold`
        :type threshold: float
        """
        super(softplusActivation, self).__init__()
        self.beta = beta
        self.threshold = threshold

    def forward(self, x, do_gradient=False, do_Hessian=False, forward_mode=True):
        r"""
        Activates each entry of incoming data via

        .. math::

            \sigma(x)  = \frac{1}{\beta}\ln(1 + e^{\beta x})
        """

        (dsigma, d2sigma) = (None, None)

        # forward propagate
        sigma = F.softplus(x, beta=self.beta, threshold=self.threshold)

        # compute derivatives
        if do_gradient or do_Hessian:
            if forward_mode is not None:
                dsigma, d2sigma = self.compute_derivatives(x, do_Hessian=do_Hessian)
            else:
                # backward mode, but do not compute yet
                self.ctx = (x,)

        return sigma, dsigma, d2sigma

    def compute_derivatives(self, *args, do_Hessian=False):
        r"""
        Computes the first and second derivatives of each entry of the incoming data via

        .. math::
            \begin{align}
                \sigma'(x)  &= \frac{1}{1 + e^{-\beta x}}\\
                \sigma''(x) &= \frac{\beta}{2\cosh(\beta x) + 2}
            \end{align}

        """
        x = args[0]
        d2sigma = None

        dsigma = 1 / (1 + torch.exp(-self.beta * x))
        if do_Hessian:
            d2sigma = self.beta / (2 + 2 * torch.cosh(self.beta * x))

        # for numerical stability
        idx = (self.beta * x > self.threshold).nonzero(as_tuple=True)
        if len(idx[0]) > 0:
            dsigma[idx] = 1.0
            if do_Hessian:
                d2sigma[idx] = 0.0

        return dsigma, d2sigma


if __name__ == '__main__':
    from hessQuik.utils import input_derivative_check
    torch.set_default_dtype(torch.float64)

    nex = 11  # no. of examples
    d = 4  # no. of input features

    x = 100 * torch.randn(nex, d)

    f = softplusActivation()

    print('======= FORWARD =======')
    input_derivative_check(f, x, do_Hessian=True, verbose=True, forward_mode=True)

    print('======= BACKWARD =======')
    input_derivative_check(f, x, do_Hessian=True, verbose=True, forward_mode=False)


import torch
import torch.nn as nn
import math
from hessQuik.layers import hessQuikLayer
import hessQuik.activations as act


class softmaxLayer(hessQuikLayer):
    """
    Forward propagation through single layer of the form

        f(u(x)) = act(u(x) @ K + b).

    Here, u(x) is the input into the layer and x is the input into the network of shapes

        x : (N, d) torch.Tensor
        u(x) : (N, in_features) torch.Tensor
        f(u(x)) : (N, out_features) torch.Tensor

    where N is the number of examples and d is the number of input features into the network.
    """

    def __init__(self, in_features):
        super(softmaxLayer, self).__init__()
        self.ctx = None
        self.in_features = in_features

    def dim_input(self):
        return self.in_features

    def dim_output(self):
        return self.in_features

    def forward(self, x, do_gradient=False, do_Hessian=False, forward_mode=True, dudx=None, d2ud2x=None):

        (dsigma, d2sigma) = (None, None)

        # forward propagate
        # sigma = torch.softmax(x, dim=1)
        sigma = x / x.sum(dim=1, keepdim=True)
        # sigma = x

        # compute derivatves
        if do_gradient or do_Hessian:
            if forward_mode is not None:
                dsigma, d2sigma = self.compute_derivatives(x, do_Hessian=do_Hessian)
            else:
                self.ctx = (x,)

        return sigma, dsigma, d2sigma

    def backward(self, *args, do_Hessian: bool = False):
        dsigma, d2sigma = self.compute_derivatives(*self.ctx, do_Hessian=do_Hessian)

        return dsigma, d2sigma

    def compute_derivatives(self, *args, do_Hessian=False):
        x = args[0]

        # exp_x = torch.exp(x)
        exp_x = x
        inv_sum_exp_x = 1.0 / exp_x.sum(dim=1, keepdim=True)
        dsigma = inv_sum_exp_x.unsqueeze(-1) * torch.eye(x.shape[1], device=x.device, dtype=x.dtype)
        dsigma -= (inv_sum_exp_x ** 2).unsqueeze(-1) * (torch.ones_like(exp_x).unsqueeze(-1) * exp_x.unsqueeze(1))
        # dsigma = torch.ones_like(x).unsqueeze(-1) * torch.eye(x.shape[1], device=x.device, dtype=x.dtype)

        d2sigma = None
        if do_Hessian:
            d2sigma = torch.zeros(x.shape[0], x.shape[1], x.shape[1], 1, device=x.device, dtype=x.dtype)

        return dsigma, d2sigma

    def extra_repr(self) -> str:
        return 'in_features={}'.format(
            self.in_features
        )


if __name__ == '__main__':
    from hessQuik.utils import input_derivative_check
    torch.set_default_dtype(torch.float64)

    nex = 11  # no. of examples
    d = 4  # no. of input features
    x = torch.randn(nex, d)

    f = softmaxLayer(d)

    print('======= FORWARD =======')
    input_derivative_check(f, x, do_Hessian=False, verbose=True, forward_mode=True)

    # print('======= BACKWARD =======')
    # input_derivative_check(f, x, do_Hessian=True, verbose=True, forward_mode=False)



import torch
import torch.nn as nn
import torch.nn.functional as F
import math
from hessQuik.layers import hessQuikLayer
from typing import Union, Tuple


class quadraticICNNLayer(hessQuikLayer):
    r"""
    Evaluate and compute derivatives of a ICNN quadratic layer.

    Examples::

        >>> import hessQuik.layers as lay
        >>> f = lay.quadraticICNNLayer(4, None, 2)
        >>> x = torch.randn(10, 4)
        >>> fx, dfdx, d2fd2x = f(x, do_gradient=True, do_Hessian=True)
        >>> print(fx.shape, dfdx.shape, d2fd2x.shape)
        torch.Size([10, 1]) torch.Size([10, 4, 1]) torch.Size([10, 4, 4, 1])

    """

    def __init__(self, input_dim: int, in_features: Union[int, None], rank: int, device=None, dtype=None) -> None:
        r"""

        :param input_dim: dimension of network inputs
        :type input_dim: int
        :param in_features: number of input features, :math:`n_{in}`.  For only ICNN quadratic layer, set ``in_features = None``
        :type in_features: int or ``None``
        :param rank: number of columns of quadratic matrix, :math:`r`.  In practice, :math:`r < n_{in}`
        :type rank: int
        :var v: weight vector for network inputs of size :math:`(d,)`
        :var w: weight vector for input features of size :math:`(n_{in},)`
        :var A: weight matrix for quadratic term of size :math:`(d, r)`
        :var mu: additive scalar bias
        :var nonneg: pointwise function to force :math:`l` to have nonnegative weights. Default ``torch.nn.functional.softplus``
        """
        factory_kwargs = {'device': device, 'dtype': dtype}
        super(quadraticICNNLayer, self).__init__()

        self.input_dim = input_dim
        self.in_features = in_features
        self.rank = rank
        self.ctx = None
        self.nonneg = F.softplus

        # create final layer
        if in_features is not None:
            self.w = nn.Parameter(torch.empty(in_features, **factory_kwargs))
        else:
            self.register_parameter('w', None)

        self.v = nn.Parameter(torch.empty(input_dim, **factory_kwargs))
        self.mu = nn.Parameter(torch.empty(1, **factory_kwargs))
        self.A = nn.Parameter(torch.empty(rank, input_dim, **factory_kwargs))
        self.reset_parameters()

    def reset_parameters(self) -> None:

        if self.in_features is not None:
            bound = 1 / math.sqrt(self.in_features)
            nn.init.uniform_(self.w, a=-bound, b=bound)
        else:
            bound = 1 / math.sqrt(self.input_dim)

        nn.init.uniform_(self.v, a=-bound, b=bound)
        nn.init.uniform_(self.mu)
        bound = 1 / math.sqrt(self.input_dim)
        nn.init.uniform_(self.A, a=-bound, b=bound)

    def dim_input(self) -> int:
        r"""
        number of input features + dimension of network inputs
        """
        return self.in_features + self.input_dim

    def dim_output(self) -> int:
        r"""
        scalar
        """
        return 1

    def forward(self, ux: torch.Tensor, do_gradient: bool = False, do_Hessian: bool = False, forward_mode: bool = True,
                dudx: Union[torch.Tensor, None] = None, d2ud2x: Union[torch.Tensor, None] = None) \
            -> Tuple[torch.Tensor, Union[torch.Tensor, None], Union[torch.Tensor, None]]:
        r"""
        Forward propagation through ICNN layer of the form, for one sample :math:`n_s = 1`,

        .. math::

            f(x) =
            \left[\begin{array}{c}u(x) & x\end{array}\right]
            \left[\begin{array}{c}w^+ \\ v\end{array}\right] + \frac{1}{2} x  A A^\top  x^\top + \mu

        Here, :math:`u(x)` is the input into the layer of size :math:`(n_s, n_{in})` which is
        a function of the input of the network, :math:`x` of size :math:`(n_s, d)`.
        The output features, :math:`f(x)`, are of size :math:`(n_s, 1)`.
        The notation :math:`(\cdot)^+` is a function that makes the weights of a matrix nonnegative.

        As an example, for one sample, :math:`n_s = 1`, the gradient with respect to :math:`x` is of the form

        .. math::

                \nabla_x f = \left[\begin{array}{c}(w^+)^\top & v^\top\end{array}\right]
                \left[\begin{array}{c} \nabla_x u \\ I\end{array}\right] + x A A^\top

        where :math:`I` is the :math:`d \times d` identity matrix.

        """
        (df, d2f) = (None, None)
        AtA = self.A.t() @ self.A

        if self.w is None:
            w = torch.empty(0, dtype=self.v.dtype, device=self.v.device)
        else:
            w = self.nonneg(self.w)

        wv = torch.cat((w, self.v), dim=0)
        x = ux[:, -self.input_dim:]

        # forward propagate
        f = ux @ wv + 0.5 * torch.sum((x @ AtA) * x, dim=1) + self.mu

        # ------------------------------------------------------------------------------------------------------------ #
        if (do_gradient or do_Hessian) and forward_mode is True:
            if self.in_features is None:
                z = torch.empty(ux.shape[0], 0)
            else:
                z = torch.zeros(ux.shape[0], self.in_features)

            df = wv.unsqueeze(0) + torch.cat((z, x @ AtA), dim=1)

            # -------------------------------------------------------------------------------------------------------- #
            if do_Hessian:
                d2f = AtA

                if d2ud2x is not None:
                    d2f = dudx[:, :, -self.input_dim:] @ d2f @ dudx[:, :, -self.input_dim:].permute(0, 2, 1)
                    z = torch.zeros(x.shape[0], self.in_features)
                    d2f += (d2ud2x @ (torch.cat((w, self.v), dim=0).unsqueeze(0)
                                      + torch.cat((z, x @ AtA), dim=1)).unsqueeze(1).unsqueeze(-1)).squeeze()

                d2f = d2f.unsqueeze(-1)
                if d2f.ndim < 4:
                    e = torch.ones(x.shape[0], device=x.device, dtype=x.dtype).view(-1, 1, 1, 1)
                    d2f = e * d2f.unsqueeze(0)

            # -------------------------------------------------------------------------------------------------------- #
            # finish computing gradient
            if dudx is not None:
                df = (dudx @ df.unsqueeze(-1)).squeeze()

            df = df.unsqueeze(-1)

        if (do_gradient or do_Hessian) and forward_mode is not True:
            self.ctx = (ux,)
            if forward_mode is False:
                df, d2f = self.backward(do_Hessian=do_Hessian)

        return f.unsqueeze(-1), df, d2f

    def backward(self, do_Hessian: bool = False,
                 dgdf: Union[torch.Tensor, None] = None, d2gd2f: Union[torch.Tensor, None] = None) \
            -> Tuple[torch.Tensor, Union[torch.Tensor, None]]:
        r"""
        Backward propagation through quadratic ICNN layer of the form, for one sample :math:`n_s = 1`,

        .. math::

                f\left(\begin{bmatrix} u & x \end{bmatrix}\right) =\left[\begin{array}{c}u & x\end{array}\right]
                \left[\begin{array}{c}w^+ \\ v\end{array}\right] + \frac{1}{2} x  A A^\top x^\top + \mu

        Here, the network is :math:`g` is a function of :math:`f(u)`.

        The gradient of the layer with respect to :math:`\begin{bmatrix} u & x \end{bmatrix}` is of the form

        .. math::

            \nabla_{[u,x]} f = \begin{bmatrix}(w^+)^\top & v^\top + x A A^\top\end{bmatrix}.

        """
        d2f = None

        ux = self.ctx[0]
        x = ux[:, -self.input_dim:]
        AtA = self.A.t() @ self.A  # TODO: recompute this or store it?

        wv = self.v
        if self.w is not None:
            wv = torch.cat((self.nonneg(self.w), wv), dim=0)

        z = torch.empty(ux.shape[0], 0)
        if self.in_features is not None:
            z = torch.zeros(ux.shape[0], self.in_features)

        df = wv.unsqueeze(0) + torch.cat((z, x @ AtA), dim=1)

        if do_Hessian:
            e = torch.ones(x.shape[0], 1, 1, dtype=AtA.dtype, device=AtA.device)
            d2f = torch.zeros(x.shape[0], ux.shape[1], ux.shape[1])
            d2f[:, -self.input_dim:, -self.input_dim:] = e * AtA
            d2f = d2f.unsqueeze(-1)

        return df.unsqueeze(-1), d2f


if __name__ == '__main__':
    from hessQuik.utils import input_derivative_check
    torch.set_default_dtype(torch.float64)

    nex = 11  # no. of examples
    d = 4  # no. of input features
    in_feat = 5
    m = 13  # rank
    x = torch.randn(nex, d)
    f = quadraticICNNLayer(d, None, m)

    print('======= FORWARD =======')
    input_derivative_check(f, x, do_Hessian=True, verbose=True, forward_mode=True)

    print('======= BACKWARD =======')
    input_derivative_check(f, x, do_Hessian=True, verbose=True, forward_mode=False)


import torch
from hessQuik.activations import hessQuikActivationFunction


class identityActivation(hessQuikActivationFunction):
    r"""
    Applies the identity activation function to each entry of the incoming data.

    Examples::

        >>> import hessQuik.activations as act
        >>> act_func = act.identityActivation()
        >>> x = torch.randn(10, 4)
        >>> sigma, dsigma, d2sigma = act_func(x, do_gradient=True, do_Hessian=True)

    """

    def __init__(self):
        super(identityActivation, self).__init__()

    def forward(self, x, do_gradient=False, do_Hessian=False, forward_mode=True):
        r"""
        Activates each entry of incoming data via

        .. math::

            \sigma(x)  = x
        """
        (dsigma, d2sigma) = (None, None)

        # forward propagate
        sigma = x

        # compute derivatives
        if do_gradient or do_Hessian:
            if forward_mode is not None:
                dsigma, d2sigma = self.compute_derivatives(x, do_Hessian=do_Hessian)
            else:
                self.ctx = (x,)

        return sigma, dsigma, d2sigma

    def compute_derivatives(self, *args, do_Hessian=False):
        r"""
        Computes the first and second derivatives of each entry of the incoming data via

        .. math::
            \begin{align}
                \sigma'(x)  &= 1\\
                \sigma''(x) &= 0
            \end{align}

        """
        x = args[0]
        d2sigma = None
        dsigma = torch.ones_like(x)

        if do_Hessian:
            d2sigma = torch.zeros_like(x)

        return dsigma, d2sigma


if __name__ == '__main__':
    from hessQuik.utils import input_derivative_check
    torch.set_default_dtype(torch.float64)

    nex = 11  # no. of examples
    d = 4  # no. of input features

    x = torch.randn(nex, d)

    f = identityActivation()

    print('======= FORWARD =======')
    input_derivative_check(f, x, do_Hessian=True, verbose=True, forward_mode=True)

    print('======= BACKWARD =======')
    input_derivative_check(f, x, do_Hessian=True, verbose=True, forward_mode=False)




import torch
from torch import Tensor
from torch.nn import Module
from typing import Tuple, List
from math import log2, floor


def module_getattr(obj: Module, names: Tuple or List or str):
    r"""
    Get specific attribute of module at any level
    """
    if isinstance(names, str) or len(names) == 1:
        if len(names) == 1:
            names = names[0]

        return getattr(obj, names)
    else:
        return module_getattr(getattr(obj, names[0]), names[1:])


def module_setattr(obj: Module, names: Tuple or List, val: Tensor):
    r"""
    Set specific attribute of module at any level
    """
    if isinstance(names, str) or len(names) == 1:
        if len(names) == 1:
            names = names[0]

        return setattr(obj, names, val)
    else:
        return module_setattr(getattr(obj, names[0]), names[1:], val)


def extract_data(net: Module, attr: str = 'data') -> (Tensor, Tuple, Tuple):
    """
    Extract data stored in specific attribute and store as 1D array
    """
    theta = torch.empty(0)
    for name, w in net.named_parameters():
        if getattr(w, attr) is None:
            w = torch.zeros_like(w.data)
        else:
            w = getattr(w, attr)

        theta = torch.cat((theta, w.reshape(-1)))

    return theta


def insert_data(net: Module, theta: Tensor) -> None:
    """
    Insert 1D array of data into specific attribute
    """
    count = 0
    for name, w in net.named_parameters():
        name_split = name.split('.')
        n = w.numel()
        module_setattr(net, name_split + ['data'], theta[count:count + n].reshape(w.shape))
        count += n


def convert_to_base(a: tuple, b: float = 2.0) -> tuple:
    """
    Convert tuple of floats to a base-exponent pair for nice printouts.

    See use in, e.g., :py:func:`hessQuik.utils.input_derivative_check.input_derivative_check`.
    """
    outputs = ()
    for i in range(len(a)):
        if a[i] <= 0:
            # catch case when equal to 0
            c, d = -1, 0
        else:
            d = floor(log2(a[i]) / log2(b))
            c = b ** (log2(a[i]) / log2(b) - d)

        outputs += (c, d)

    return outputs



from hessQuik.networks import NN
from hessQuik.layers import singleLayer
import hessQuik.activations as act
from typing import Union, Tuple, List
from copy import deepcopy


class fullyConnectedNN(NN):
    r"""
    Fully-connected network where every layer is a single layer.  Let :math:`u_0 = x` be the input into the network.
    The construction is of the form

    .. math::

        \begin{align}
            u_1 &= \sigma(K_1 u_0 + b_1)\\
            u_2 &= \sigma(K_2 u_1 + b_2)\\
                &\vdots \\
            u_{\ell} &= \sigma(K_{\ell} u_{\ell-1} + b_{\ell})
        \end{align}

    where :math:`\ell` is the number of layers.
    Each vector of features :math:`u_i` is of size :math:`(n_s, n_i)` where :math:`n_s` is the number of samples
    and :math:`n_i` is the dimension or width of the hidden features on layer :math:`i`.
    Users choose the widths of the network and the activation function :math:`\sigma`.
    """

    def __init__(self, widths: Union[Tuple, List], act: act.hessQuikActivationFunction = act.softplusActivation(),
                 device=None, dtype=None):
        r"""
        :param widths: dimension of hidden features
        :type widths: tuple or list
        :param act: hessQuik activation function. Default: hessQuik.activations.softplusActivationFunction
        :type act: hessQuik.activations.hessQuikActivationFunction
        """
        factory_kwargs = {'device': device, 'dtype': dtype}
        args = ()
        for i in range(len(widths) - 1):
            args += (singleLayer(widths[i], widths[i + 1], act=deepcopy(act), **factory_kwargs),)

        super(fullyConnectedNN, self).__init__(*args)


if __name__ == '__main__':
    import torch
    from hessQuik.utils import input_derivative_check
    torch.set_default_dtype(torch.float64)

    # problem setup
    nex = 11
    d = 3
    x = torch.randn(nex, d)
    f = fullyConnectedNN([d, 2, 5, 1], act=act.softplusActivation())

    print('======= FORWARD =======')
    input_derivative_check(f, x, do_Hessian=True, verbose=True, forward_mode=True)

    print('======= BACKWARD =======')
    input_derivative_check(f, x, do_Hessian=True, verbose=True, forward_mode=False)

    # widths1 = [2, 3]
    # widths2 = [4, 5]
    # widths3 = [7, 6, 2]
    #
    # f = NN(fullyConnectedNN([d] + widths1, act=act.antiTanhActivation()),
    #        fullyConnectedNN([widths1[-1]] + widths2, act=act.identityActivation()),
    #        fullyConnectedNN([widths2[-1]] + widths3, act=act.softplusActivation()))
    #
    # print('======= FORWARD NN =======')
    # input_derivative_check(f, x, do_Hessian=True, verbose=True, forward_mode=True)
    #
    # print('======= BACKWARD NN =======')
    # input_derivative_check(f, x, do_Hessian=True, verbose=True, forward_mode=False)



import torch.nn as nn
from torch import Tensor
from typing import Union, Tuple


class hessQuikActivationFunction(nn.Module):
    r"""
    Base class for all hessQuik activation functions.
    """

    def __init__(self) -> None:
        super(hessQuikActivationFunction, self).__init__()
        self.ctx = None  # context variable

    def forward(self, x: Tensor, do_gradient: bool = False, do_Hessian: bool = False, forward_mode: bool = True) \
            -> Tuple[Tensor, Union[Tensor, None], Union[Tensor, None]]:
        r"""
        Applies a pointwise activation function to the incoming data.

        :param x: input into the activation function. :math:`(*)` where :math:`*` means any shape.
        :type x: torch.Tensor
        :param do_gradient: If set to ``True``, the gradient will be computed during the forward call. Default: ``False``
        :type do_gradient: bool, optional
        :param do_Hessian: If set to ``True``, the Hessian will be computed during the forward call. Default: ``False``
        :type do_Hessian: bool, optional
        :param forward_mode:  If set to ``False``, the derivatives will be computed in backward mode. Default: ``True``
        :type forward_mode: bool, optional
        :return:

            - **sigma** (*torch.Tensor*) - value of activation function at input x, same size as x
            - **dsigma** (*torch.Tensor* or ``None``) - first derivative of activation function at input x, same size as x
            - **d2sigma** (*torch.Tensor* or ``None``) - second derivative of activation function at input x, same size as x

        """
        raise NotImplementedError

    def backward(self, do_Hessian: bool = False) -> Tuple[Tensor, Union[Tensor, None]]:
        r"""
        Computes derivatives of activation function evaluated at x in backward mode.

        Calls self.compute_derivatives without inputs, stores necessary variables in self.ctx.

        Inherited by all subclasses.
        """
        dsigma, d2sigma = self.compute_derivatives(*self.ctx, do_Hessian=do_Hessian)

        return dsigma, d2sigma

    def compute_derivatives(self, *args, do_Hessian: bool = False) -> Tuple[Tensor, Union[Tensor, None]]:
        r"""

        :param args: variables needed to compute derivatives
        :type args: torch.Tensor
        :param do_Hessian: If set to ``True``, the Hessian will be computed during the forward call. Default: ``False``
        :type do_Hessian: bool, optional
        :return:

            - **dsigma** (*torch.Tensor* or ``None``) - first derivative of activation function at input x, same size as x
            - **d2sigma** (*torch.Tensor* or ``None``) - second derivative of activation function at input x, same size as x

        """
        raise NotImplementedError



import torch
from typing import Union


def train_one_epoch(f: torch.nn.Module, x: torch.Tensor, y: torch.Tensor, optimizer: torch.optim.Optimizer,
                    batch_size: int = 5, do_gradient: bool = True, do_Hessian: bool = True,
                    loss_weights: Union[tuple, list] = (1.0, 1.0, 1.0)):
    r"""
    Training mean-square loss for one epoch where the loss function is

    .. math::
        L(\theta) = \frac{1}{2N}\left(w_0\|f_{\theta}(x)\|^2 + w_1\|\nabla f_{\theta}(x)\|^2 + w_2\|\nabla^2 f_{\theta}(x)\|^2\right)

    where :math:`f_{theta}` is the network :math:`\theta` are the network weights,
    and :math:`N` is the number of training samples.
    The loss corresponding to the function value, gradient, and Hessian each can have different weights,
    :math:`w_0`, :math:`w_1`, and :math:`w_2`, respectively.

    :param f: hessQuik neural network to train
    :type f: torch.nn.Module
    :param x: training data of shape :math:`(N, *)` where :math:`*` can be any shapw
    :type x: torch.Tensor
    :param y: target data of shape :math:`(N, *)`
    :type y: torch.Tensor
    :param optimizer: method for updating the network weights
    :type optimizer: torch.optim.Optimizer
    :param batch_size: size of mini-batches for stochatistic training. Default: 5
    :type batch_size: int
    :param do_gradient: If set to ``True``, the gradient will be computed during the forward call. Default: ``False``
    :type do_gradient: bool, optional
    :param do_Hessian: If set to ``True``, the Hessian will be computed during the forward call. Default: ``False``
    :type do_Hessian: bool, optional
    :param loss_weights: weight for each term in the loss function
    :type loss_weights: tuple or list
    :return: tuple containing the overall running loss and the running loss for each term in the loss function

    """
    f.train()
    n = x.shape[0]
    b = batch_size
    n_batch = n // b

    (loss_f, loss_df, loss_d2f) = (torch.zeros(1), torch.zeros(1), torch.zeros(1))
    running_loss, running_loss_f, running_loss_df, running_loss_d2f = (0.0, 0.0, 0.0 if do_gradient else None,
                                                                       0.0 if do_Hessian else None)

    # shuffle
    idx = torch.randperm(n)

    for i in range(n_batch):
        idxb = idx[i * b:(i + 1) * b]
        xb, yb = x[idxb], y[idxb]

        optimizer.zero_grad()
        fb, dfb, d2fb = f(xb, do_gradient=do_gradient, do_Hessian=do_Hessian)

        loss_f = (0.5 / b) * torch.norm(fb - yb[:, 0].view_as(fb)) ** 2
        loss = loss_weights[0] * loss_f
        running_loss_f += b * loss_f.item()

        if do_gradient:
            loss_df = (0.5 / b) * torch.norm(dfb - yb[:, 1:xb.shape[1]+1].view_as(dfb)) ** 2
            loss = loss + loss_weights[1] * loss_df
            running_loss_df += b * loss_df.item()
        if do_Hessian:
            loss_d2f = (0.5 / b) * torch.norm(d2fb - yb[:, xb.shape[1]+1:].view_as(d2fb)) ** 2
            loss = loss + loss_weights[2] * loss_d2f
            running_loss_d2f += b * loss_d2f.item()

        running_loss += b * loss.item()

        # update network weights
        loss.backward()
        optimizer.step()

    output = (running_loss / n, running_loss_f / n)
    if do_gradient:
        output += (running_loss_df / n,)
        if do_Hessian:
            output += (running_loss_d2f / n,)

    return output


def test(f: torch.nn.Module, x: torch.Tensor, y: torch.Tensor, do_gradient: bool = True, do_Hessian: bool = True,
         loss_weights: Union[tuple, list] = (1.0, 1.0, 1.0)):
    r"""
    Evaluate mean-squared loss function without training

    See :py:func:`hessQuik.utils.training.train_one_epoch` for details.
    """
    f.eval()

    (loss_f, loss_df, loss_d2f) = (torch.zeros(1), torch.zeros(1), torch.zeros(1))
    with torch.no_grad():
        n = x.shape[0]
        f0, df0, d2f0 = f(x, do_gradient=do_gradient, do_Hessian=do_Hessian)

        loss_f = (0.5 / n) * torch.norm(f0 - y[:, 0].view_as(f0)) ** 2
        loss = loss_weights[0] * loss_f

        if do_gradient:
            loss_df = (0.5 / n) * torch.norm(df0 - y[:, 1:x.shape[1]+1].view_as(df0)) ** 2
            loss +=  loss_weights[1] * loss_df

        if do_Hessian:
            loss_d2f = (0.5 / n) * torch.norm(d2f0 - y[:, x.shape[1]+1:].view_as(d2f0)) ** 2
            loss += loss_weights[2] *  loss_d2f

    output = (loss.item(), loss_f.item())
    if do_gradient:
        output += (loss_df.item(),)
        if do_Hessian:
            output += (loss_d2f.item(),)

    return output


def print_headers(do_gradient: bool = True, do_Hessian: bool = True, verbose: bool = True,
                  loss_weights: Union[tuple, list] = (1.0, 1.0, 1.0)):
    r"""
    Print headers for nice training
    """
    loss_printouts = ('loss', 'loss_f')
    if do_gradient or do_Hessian:
        loss_printouts += ('loss_df',)
        if do_Hessian:
            loss_printouts += ('loss_d2f',)
    n_loss = len(loss_printouts)

    headers = (('', '', '|', 'running',) + (n_loss - 1) * ('',) + ('|', 'train',)
               + (n_loss - 1) * ('',) + ('|', 'valid',) + (n_loss - 1) * ('',))
    weights = (('', '', '|', 'weights',) + loss_weights[:n_loss - 1] + ('|', '',)
               + loss_weights[:n_loss - 1] + ('|', '',) + loss_weights[:n_loss - 1])

    printouts = ('epoch', 'time') + 3 * (('|',) + loss_printouts)
    printouts_frmt = '{:<15d}{:<15.4f}' + 3 * ('{:<2s}' + n_loss * '{:<15.4e}')

    if verbose:
        print(('{:<15s}{:<15s}' + 3 * ('{:<2s}' + n_loss * '{:<15s}')).format(*headers))
        print(('{:<15s}{:<15s}' + 3 * ('{:<2s}{:<15s}' + (n_loss - 1) * '{:<15.2e}')).format(*weights))
        print(('{:<15s}{:<15s}' + 3 * ('{:<2s}' + n_loss * '{:<15s}')).format(*printouts))

    return headers, printouts, printouts_frmt



import torch
import matplotlib.pyplot as plt
import matplotlib.colors as colors
import numpy as np

import os
import pickle

plt.rcParams.update({'font.size': 32})
plt.rcParams.update({'image.interpolation': None})
plt.rcParams['figure.figsize'] = [18, 12]
plt.rcParams['figure.dpi'] = 200


#%% HESSQUIK VS PYTORCH
load_dir = '/Users/elizabethnewman/Desktop/hessQuikResults/scalar/'

os.chdir(load_dir)

markers = ['o', '^', 's']
filled = [True, False, False]
linewidth = 6
markersize = 20


plt.figure()
for i, name in enumerate(['hessQuik', 'PytorchAD', 'PytorchHessian']):
    output = pickle.load(open('12-22-2021--' + name + '-resnet-cpu-w16-d4-out1.p', 'rb'))
    results = output['results']

    x = results['in_feature_range']
    y = results['timing_trials_mean'].squeeze()

    plt.semilogy(x, y, '-' + markers[i], linewidth=linewidth, markersize=markersize, label=name + ': cpu')

    output = pickle.load(open('12-22-2021--' + name + '-resnet-cuda-w16-d4-out1.p', 'rb'))
    results = output['results']

    x = results['in_feature_range']
    y = results['timing_trials_mean'].squeeze()

    plt.semilogy(x, y, '--' + markers[i], linewidth=linewidth, markersize=markersize, label=name + ': cuda')


plt.xscale('log', base=2)
plt.yscale('log', base=10)
plt.xlabel('in features')
plt.ylabel('time (seconds)')
plt.grid()
plt.ylim(1e-3, 2e1)
plt.legend()
plt.show()

#%% WIDTH AND DEPTH
load_dir = '/Users/elizabethnewman/Desktop/hessQuikResults/scalar'
os.chdir(load_dir)
for w in [8, 16, 32, 64]:
    output = pickle.load(open('12-17-2021---hessQuik-cpu-w' + str(w) + '-d4.p', 'rb'))
    results = output['results']
    timing_trials_mean = results['timing_trials_mean']
    in_feature_range = results['in_feature_range']
    out_feature_range = results['out_feature_range']

    x = results['in_feature_range']
    y = results['timing_trials_mean'].squeeze()

    p = plt.semilogy(x, y, '-o', linewidth=linewidth, markersize=markersize, label=str(w))
    #
    # output = pickle.load(open('12-17-2021---hessQuik-cuda-w' + str(w) + '-d4.p', 'rb'))
    # results = output['results']
    # timing_trials_mean = results['timing_trials_mean']
    # in_feature_range = results['in_feature_range']
    # out_feature_range = results['out_feature_range']
    #
    # x = results['in_feature_range']
    # y = results['timing_trials_mean'].squeeze()
    #
    # plt.semilogy(x, y, '--o', linewidth=linewidth, markersize=markersize, label=str(w), color=p[0].get_color())


plt.xscale('log', base=2)
plt.yscale('log', base=10)
plt.xlabel('in features')
plt.ylabel('time (seconds)')
plt.grid()
plt.ylim(1e-3, 2e1)
plt.legend()
plt.show()


for d in [2, 4, 8, 16]:
    output = pickle.load(open('12-17-2021---hessQuik-cpu-w16-d' + str(d) + '.p', 'rb'))
    results = output['results']
    timing_trials_mean = results['timing_trials_mean']
    in_feature_range = results['in_feature_range']
    out_feature_range = results['out_feature_range']

    x = results['in_feature_range']
    y = results['timing_trials_mean'].squeeze()
    #
    p = plt.semilogy(x, y, '-o', linewidth=linewidth, markersize=markersize, label=str(d))
    #
    # output = pickle.load(open('12-17-2021---hessQuik-cuda-w16-d' + str(d) + '.p', 'rb'))
    # results = output['results']
    # timing_trials_mean = results['timing_trials_mean']
    # in_feature_range = results['in_feature_range']
    # out_feature_range = results['out_feature_range']
    #
    # x = results['in_feature_range']
    # y = results['timing_trials_mean'].squeeze()
    #
    # plt.semilogy(x, y, '--p', linewidth=linewidth, markersize=markersize, label=str(d), color=p[0].get_color())

plt.xscale('log', base=2)
plt.yscale('log', base=10)
plt.xlabel('in features')
plt.ylabel('time (seconds)')
plt.grid()
plt.ylim(1e-3, 2e1)
plt.legend()
plt.show()

#%% VECTOR OUTPUT
load_dir = '/Users/elizabethnewman/Desktop/hessQuikResults/vector/'

plt.rcParams.update({'font.size': 24})
plt.rcParams.update({'image.interpolation': None})
plt.rcParams['figure.figsize'] = [14, 12]
plt.rcParams['figure.dpi'] = 200

os.chdir(load_dir)

fig, axes = plt.subplots(nrows=1, ncols=4)
for i, name in enumerate(['cpu', 'cuda']):
    output = pickle.load(open('12-22-2021--hessQuik-resnet-' + name + '-w16-d4-out4.p', 'rb'))
    results = output['results']
    timing_trials_mean = results['timing_trials_mean']
    in_feature_range = results['in_feature_range']
    out_feature_range = results['out_feature_range']
    print(timing_trials_mean)

    # plt.subplot(1, 4, 2 * i + 1)

    im = axes[2 * i].imshow(torch.flipud(timing_trials_mean), norm=colors.LogNorm(vmin=1e-3, vmax=1e2))
    plt.sca(axes[2 * i])
    plt.xticks(list(torch.arange(len(out_feature_range)).numpy()), labels=['$2^0$', '$2^1$', '$2^2$', '$2^3$'])
    # plt.xlabel('out_features')

    if i == 0:
        plt.yticks(list(torch.arange(len(in_feature_range)).numpy()),
                   labels=['$2^{10}$', '$2^9$', '$2^8$', '$2^7$', '$2^6$', '$2^5$', '$2^4$', '$2^3$', '$2^2$', '$2^1$', '$2^0$'])
    else:
        plt.tick_params(axis='y', left=False, right=False, labelleft=False)
    plt.title('hessQuik: ' + name, fontsize=16)

    output = pickle.load(open('12-22-2021--PytorchAD-resnet-' + name + '-w16-d4-out4.p', 'rb'))
    results = output['results']
    timing_trials_mean = results['timing_trials_mean']
    in_feature_range = results['in_feature_range']
    out_feature_range = results['out_feature_range']
    print(timing_trials_mean)


    im = axes[2 * i + 1].imshow(torch.flipud(timing_trials_mean), norm=colors.LogNorm(vmin=1e-3, vmax=1e2))
    plt.sca(axes[2 * i + 1])
    plt.xticks(list(torch.arange(len(out_feature_range)).numpy()), labels=['$2^0$', '$2^1$', '$2^2$', '$2^3$'])
    plt.tick_params(axis='y', left=False, right=False, labelleft=False)
    plt.title('PytorchAD: ' + name, fontsize=16)

fig.subplots_adjust(right=0.8)
cbar_ax = fig.add_axes([0.85, 0.27, 0.02, 0.45])
fig.colorbar(im, cax=cbar_ax)
fig.text(0.02, 0.5, 'input features', va='center', rotation='vertical')
fig.text(0.5, 0.15, 'output features', ha='center')
plt.show()




import torch
import torch.nn as nn
import math
from hessQuik.layers import hessQuikLayer
import hessQuik.activations as act
from typing import Union, Tuple

class singleLayer(hessQuikLayer):
    r"""
    Evaluate and compute derivatives of a single layer.

    Examples::

        >>> import hessQuik.layers as lay
        >>> f = lay.singleLayer(4, 7)
        >>> x = torch.randn(10, 4)
        >>> fx, dfdx, d2fd2x = f(x, do_gradient=True, do_Hessian=True)
        >>> print(fx.shape, dfdx.shape, d2fd2x.shape)
        torch.Size([10, 7]) torch.Size([10, 4, 7]) torch.Size([10, 4, 4, 7])

    """

    def __init__(self, in_features: int, out_features: int,
                 act: act.hessQuikActivationFunction = act.identityActivation(),
                 device=None, dtype=None) -> None:
        r"""
        :param in_features: number of input features, :math:`n_{in}`
        :type in_features: int
        :param out_features: number of output features, :math:`n_{out}`
        :type out_features: int
        :param act: activation function
        :type act: hessQuikActivationFunction
        :var K: weight matrix of size :math:`(n_{in}, n_{out})`
        :var b: bias vector of size :math:`(n_{out},)`
        """
        factory_kwargs = {'device': device, 'dtype': dtype}
        super(singleLayer, self).__init__()

        self.in_features = in_features
        self.out_features = out_features
        self.act = act

        self.K = nn.Parameter(torch.empty(in_features, out_features, **factory_kwargs))
        self.b = nn.Parameter(torch.empty(out_features, **factory_kwargs))
        self.reset_parameters()

    def reset_parameters(self) -> None:
        nn.init.kaiming_uniform_(self.K, a=math.sqrt(5))

        fan_in, _ = nn.init._calculate_fan_in_and_fan_out(self.K)
        bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
        nn.init.uniform_(self.b, -bound, bound)

    def dim_input(self) -> int:
        r"""
        number of input features
        """
        return self.in_features

    def dim_output(self):
        r"""
        number of output features
        """
        return self.out_features

    def forward(self, u: torch.Tensor, do_gradient: bool = False, do_Hessian: bool = False, forward_mode: bool = True,
                dudx: Union[torch.Tensor, None] = None, d2ud2x: Union[torch.Tensor, None] = None) \
            -> Tuple[torch.Tensor, Union[torch.Tensor, None], Union[torch.Tensor, None]]:
        r"""
        Forward propagation through single layer of the form

        .. math::

            f(x) = \sigma(u(x) K + b)

        Here, :math:`u(x)` is the input into the layer of size :math:`(n_s, n_{in})` which is
        a function of the input of the network, :math:`x`.
        The output features, :math:`f(x)`, are of size :math:`(n_s, n_{out})`.

        As an example, for one sample, :math:`n_s = 1`, the gradient with respect to :math:`x` is of the form

        .. math::

            \nabla_x f = \text{diag}(\sigma'(u(x) K + b))K^\top \nabla_x u

        where :math:`\text{diag}` transforms a vector into the entries of a diagonal matrix.
        """
        (dfdx, d2fd2x) = (None, None)
        f, dsig, d2sig = self.act(u @ self.K + self.b, do_gradient=do_gradient, do_Hessian=do_Hessian,
                                  forward_mode=True if forward_mode is True else None)
        # ------------------------------------------------------------------------------------------------------------ #
        # forward mode
        if (do_gradient or do_Hessian) and forward_mode is True:
            dfdx = dsig.unsqueeze(1) * self.K
            # -------------------------------------------------------------------------------------------------------- #
            if do_Hessian:
                d2fd2x = (d2sig.unsqueeze(1) * self.K).unsqueeze(2) * self.K.unsqueeze(0).unsqueeze(0)

                # Gauss-Newton approximation
                if d2ud2x is not None:
                    d2fd2x = dudx.unsqueeze(1) @ (d2fd2x.permute(0, 3, 1, 2) @ dudx.permute(0, 2, 1).unsqueeze(1))
                    d2fd2x = d2fd2x.permute(0, 2, 3, 1)

                    # extra term to compute full Hessian
                    d2fd2x += d2ud2x @ dfdx.unsqueeze(1)
            # -------------------------------------------------------------------------------------------------------- #
            # finish computing gradient
            if dudx is not None:
                dfdx = dudx @ dfdx

        # ------------------------------------------------------------------------------------------------------------ #
        # backward mode (if layer is not wrapped in NN)
        if (do_gradient or do_Hessian) and forward_mode is False:
            dfdx, d2fd2x = self.backward(do_Hessian=do_Hessian)

        return f, dfdx, d2fd2x

    def backward(self, do_Hessian: bool = False,
                 dgdf: Union[torch.Tensor, None] = None, d2gd2f: Union[torch.Tensor, None] = None):
        r"""
        Backward propagation through single layer of the form

        .. math::

                f(u) = \sigma(u K + b)

        Here, the network is :math:`g` is a function of :math:`f(u)`.

        As an example, for one sample, :math:`n_s = 1`, the gradient of the network with respect to :math:`u` is of the form

        .. math::

            \nabla_u g = (\sigma'(u K + b) \odot \nabla_f g)K^\top

        where :math:`\odot` denotes the pointwise product.

        """

        d2gd2u = None
        dsig, d2sig = self.act.backward(do_Hessian=do_Hessian)
        dgdu = dsig.unsqueeze(1) * self.K

        if do_Hessian:
            d2gd2u = (d2sig.unsqueeze(1) * self.K.unsqueeze(0)).unsqueeze(2) * self.K.unsqueeze(0).unsqueeze(0)

            if d2gd2f is not None:
                # Gauss-Newton approximation
                h1 = (dgdu.unsqueeze(1) @ d2gd2f.permute(0, 3, 1, 2) @ dgdu.permute(0, 2, 1).unsqueeze(1))
                h1 = h1.permute(0, 2, 3, 1)

                # extra term to compute full Hessian
                h2 = d2gd2u @ dgdf.unsqueeze(1)

                # combine
                d2gd2u = h1 + h2

        # finish computing gradient
        if dgdf is not None:
            dgdu = dgdu @ dgdf

        return dgdu, d2gd2u

    def extra_repr(self) -> str:
        r"""
        :meta private:
        """
        return 'in_features={}, out_features={}'.format(
            self.in_features, self.out_features
        )


if __name__ == '__main__':
    from hessQuik.utils import input_derivative_check
    torch.set_default_dtype(torch.float64)

    nex = 11  # no. of examples
    d = 4  # no. of input features
    m = 7  # no. of output features
    x = torch.randn(nex, d)

    f = singleLayer(d, m, act=act.softplusActivation())

    print('======= FORWARD =======')
    input_derivative_check(f, x, do_Hessian=True, verbose=True, forward_mode=True)

    print('======= BACKWARD =======')
    input_derivative_check(f, x, do_Hessian=True, verbose=True, forward_mode=False)



import torch
import hessQuik
import hessQuik.activations as act
import hessQuik.layers as lay
import hessQuik.networks as net
import time
import gc
from typing import Union


def setup_device_and_gradient(f: hessQuik.networks.NN, network_wrapper: str = 'hessQuik', device: str = 'cpu') \
        -> torch.nn.Module:
    r"""
    Setup network with correct wrapper and device
    """
    # map to device
    f = f.to(device)

    if network_wrapper == 'PytorchAD':
        f = net.NNPytorchAD(f)

    if network_wrapper == 'PytorchHessian':
        f = net.NNPytorchHessian(f)

    return f


def setup_resnet(in_features: int, out_features: int, width: int = 16, depth: int = 4) -> hessQuik.networks.NN:
    r"""
    Setup resnet architecture for timing tests
    """

    f = net.NN(lay.singleLayer(in_features, width, act=act.antiTanhActivation()),
               net.resnetNN(width, depth, h=0.5, act=act.tanhActivation()),
               lay.singleLayer(width, out_features, act=act.identityActivation()))
    return f


def setup_fully_connected(in_features: int, out_features: int, width: int = 16, depth: int = 4) -> hessQuik.networks.NN:
    r"""
    Setup fully-connected architecture for timing tests
    """
    f = net.fullyConnectedNN([in_features] + depth * [width] + [out_features], act=act.tanhActivation())
    return f


def setup_icnn(in_features: int, out_features: int, width: int = 16, depth: int = 4) -> hessQuik.networks.NN:
    r"""
    Setup ICNN architecture for timing tests.

    Requires scalar output.
    """

    if out_features > 1:
        raise ValueError('Must be scalar output for ICNN example')

    f = net.NN(net.ICNN(in_features, [None] + depth * [width], act=act.tanhActivation()),
               lay.quadraticICNNLayer(in_features, width, rank=2))
    return f


def setup_network(in_features: int, out_features: int, width: int, depth: int, network_type: str = 'resnet',
                  network_wrapper: str = 'hessQuik', device: str = 'cpu'):
    r"""
    Wrapper to setup network.
    """
    if network_type == 'resnet':
        f = setup_resnet(in_features, out_features, width, depth)
    elif network_type == 'fully_connected':
        f = setup_fully_connected(in_features, out_features, width, depth)
    elif network_type == 'icnn':
        f = setup_icnn(in_features, out_features, width, depth)
    else:
        raise ValueError('network_type must be "resnet", "fully_connected", or "icnn"')

    f = setup_device_and_gradient(f, network_wrapper=network_wrapper, device=device)

    return f


def timing_test_cpu(f: Union[hessQuik.networks.NN, torch.nn.Module], x: torch.Tensor, num_trials: int = 10,
                    clear_memory: bool = True) -> torch.Tensor:
    r"""
    Timing test for one architecture on CPU.

    Test is run ``num_trials`` times and the timing for each trial is returned.

    The timing includes one dry run for the first iteration that is not returned.

    Memory is cleared after the completion of all trials.
    """
    total_time = torch.zeros(num_trials + 1)
    for i in range(num_trials + 1):
        t1_start = time.perf_counter()
        f0, df0, d2f0 = f(x, do_gradient=True, do_Hessian=True)
        t1_stop = time.perf_counter()
        total_time[i] = t1_stop - t1_start

    if clear_memory:
        del f, x
        gc.collect()
        torch.cuda.empty_cache()

    return total_time[1:]


def timing_test_gpu(f: Union[hessQuik.networks.NN, torch.nn.Module], x: torch.Tensor,
                    num_trials: int = 10, clear_memory: bool = True):
    r"""
    Timing test for one architecture on CPU.

    Test is run ``num_trials`` times and the timing for each trial is returned.

    Each trial includes a ``torch.cuda.synchonize`` call.

    The timing includes one dry run for the first iteration that is not returned.

    Memory is cleared after the completion of all trials.
    """

    total_time = torch.zeros(num_trials + 1)
    for i in range(num_trials + 1):
        t1_start = time.perf_counter()
        f0, df0, d2f0 = f(x, do_gradient=True, do_Hessian=True)
        torch.cuda.synchronize()
        t1_stop = time.perf_counter()
        total_time[i] = t1_stop - t1_start

    if clear_memory:
        del f, x
        gc.collect()
        torch.cuda.empty_cache()

    return total_time[1:]


def timing_test(in_feature_range: torch.Tensor, out_feature_range: torch.Tensor, nex: int = 10, num_trials: int = 10, 
                width: int = 16, depth: int = 4, network_wrapper: str = 'hessQuik', 
                network_type: str = 'resnet', device: str = 'cpu', clear_memory: bool = True) -> dict:
    r"""
    
    :param in_feature_range: available input feature dimensions
    :type in_feature_range: torch.Tensor
    :param out_feature_range: available output feature dimensions
    :type out_feature_range: torch.Tensor
    :param nex: number of examples for network input. Default: 10
    :type nex: int, optional
    :param num_trials: number of trials per input-output feature combination. Default: 10
    :type num_trials: int, optional
    :param width: width of network. Default: 16
    :type width: int, optional
    :param depth: depth of network. Default: 4
    :type depth: int, optional
    :param network_wrapper: type of network wrapper. Default: 'hessQuik'. Options: 'hessQuik', 'PytorchAD', 'PytorchHessian'
    :type network_wrapper: str, optional
    :param network_type: network architecture. Default: 'resnet'. Options: 'resnet', 'fully_connected', 'icnn'
    :type network_type: str, optional
    :param device: device for testing. Default: 'cpu'
    :type device: str, optional
    :param clear_memory: flag to clear memory after each set of trials
    :type clear_memory: bool, optional
    :return: dictionary containing keys

        - **'timing_trials'** (*torch.Tensor*) - time (in seconds) for each trial and each architecture
        - **'timing_trials_mean'** (*torch.Tensor*) - average over trials for each architecture
        - **'timing_trials_std'** (*torch.Tensor*) - standard deviation over trials for each architecture
        - **'in_feature_range'** (*torch.Tensor*) - available input feature dimensions
        - **'out_feature_range'** (*torch.Tensor*) - available output feature dimensions
        - **'nex'** (*int*) - number of samples for the input data
        - **'num_trials'** (*int*) - number of trials per architecture

    """

    # initialize
    timing_trials = torch.zeros(len(in_feature_range), len(out_feature_range), num_trials)
    timing_trials_mean = torch.zeros(len(in_feature_range), len(out_feature_range))
    timing_trials_std = torch.zeros_like(timing_trials_mean)

    for idx_out, out_features in enumerate(out_feature_range):
        for idx_in, in_features in enumerate(in_feature_range):

            # setup network
            f = setup_network(in_features, out_features, width, depth,  network_type=network_type,
                              network_wrapper=network_wrapper, device=device)

            # setup data
            x = torch.randn(nex, in_features, device=device)

            # main test
            if device == 'cpu':
                total_time = timing_test_cpu(f, x, num_trials=num_trials, clear_memory=clear_memory)
            else:
                total_time = timing_test_gpu(f, x, num_trials=num_trials, clear_memory=clear_memory)

            # store results
            timing_trials[idx_in, idx_out] = total_time
            timing_trials_mean[idx_in, idx_out] = torch.mean(total_time).item()
            timing_trials_std[idx_in, idx_out] = torch.std(total_time).item()

    # store setup and results
    results = {'timing_trials': timing_trials,
               'timing_trials_mean': timing_trials_mean,
               'timing_trials_std': timing_trials_std,
               'in_feature_range': in_feature_range,
               'out_feature_range': out_feature_range,
               'nex': nex,
               'num_trials': num_trials}

    return results



import torch
import numpy as np
import matplotlib.pyplot as plt


def network_imshow(net, grid_x, grid_y, c_true):

    grid_data = torch.cat((grid_x.reshape(-1, 1), grid_y.reshape(-1, 1)), dim=1)

    # approximation
    c_pred = net.forward(grid_data).detach().view(grid_x.shape).numpy()

    print('relative error = ',
          np.linalg.norm(c_pred.reshape(-1) - c_true.reshape(-1)) / np.linalg.norm(c_true.reshape(-1)))

    # image plots
    fig, axs = plt.subplots(2, 2)
    ax = axs[0, 0]
    p = ax.imshow(c_true.reshape(grid_x.shape))
    ax.axis('off')
    ax.set_title('true')
    fig.colorbar(p, ax=ax, aspect=10)

    ax = axs[0, 1]
    p = ax.imshow(c_pred.reshape(grid_x.shape))
    ax.axis('off')
    ax.set_title('approx')
    fig.colorbar(p, ax=ax, aspect=10)

    ax = axs[1, 0]
    p = ax.imshow(np.abs(c_pred - c_true).reshape(grid_x.shape))
    fig.colorbar(p, ax=ax, aspect=10)
    ax.axis('off')
    ax.set_title('abs. diff.')

    ax = axs[1, 1]
    ax.axis('off')



import torch
import torch.nn as nn
from torch.autograd import grad
from torch.autograd.functional import hessian
from typing import Union, Tuple


class NN(nn.Sequential):
    r"""
    Wrapper for hessQuik networks built upon torch.nn.Sequential.
    """
    def __init__(self, *args):
        r"""
        :param args: sequence of hessQuik layers to be concatenated
        """
        # check for compatible composition
        for i, _ in enumerate(args[1:], start=1):
            n_out = args[i - 1].dim_output()
            n_in = args[i].dim_input()

            if not (n_out == n_in):
                raise ValueError("incompatible composition for block " + str(i - 1) + " to block " + str(i))

        super(NN, self).__init__(*args)

    def dim_input(self) -> int:
        r"""
        Number of network input features
        """
        return self[0].dim_input()

    def dim_output(self):
        r"""
        Number of network output features
        """
        return self[-1].dim_output()

    def setup_forward_mode(self, **kwargs):
        r"""
        Setup forward or backward mode.

        If ``kwargs`` does not include a ``forward_mode`` key, then the heuristic is to use ``forward_mode = True``
        if :math:`n_{in} < n_{out}` where :math:`n_{in}` is the number of input features and
        :math:`n_{out}` is the number of output features.

        There are three possible options once ``forward_mode`` is a key of ``kwargs``:

            - If ``forward_mode = True``, then the network computes derivatives during forward propagation.
            - If ``forward_mode = False``, then the network calls the backward routine to compute derivatives after forward propagating.
            - If ``forward_mode = None``, then the network will compute derivatives in backward mode, but will not call the backward routine.  This enables concatenation of networks, not just layers.
        """
        if not ('forward_mode' in kwargs.keys()):
            if self.dim_input() < self.dim_output():
                forward_mode = True  # compute the derivatives in forward mode
            else:
                forward_mode = False  # store necessary info, but do not compute derivatives until backward call
            kwargs['forward_mode'] = forward_mode

        return kwargs['forward_mode']

    def forward(self, x: torch.Tensor, do_gradient: bool = False, do_Hessian: bool = False,
                dudx: Union[torch.Tensor, None] = None, d2ud2x: Union[torch.Tensor, None] = None, **kwargs) \
            -> Tuple[torch.Tensor, Union[torch.Tensor, None], Union[torch.Tensor, None]]:
        r"""
        Forward propagate through network and compute derivatives

        :param x: input into network of shape :math:`(n_s, d)` where :math:`n_s` is the number of samples and :math:`d` is the number of input features
        :type x: torch.Tensor
        :param do_gradient: If set to ``True``, the gradient will be computed during the forward call. Default: ``False``
        :type do_gradient: bool, optional
        :param do_Hessian: If set to ``True``, the Hessian will be computed during the forward call. Default: ``False``
        :type do_Hessian: bool, optional
        :param dudx: if ``forward_mode = True``, gradient of features from previous layer with respect to network input :math:`x` with shape :math:`(n_s, d, n_{in})`
        :type dudx: torch.Tensor or ``None``
        :param d2ud2x: if ``forward_mode = True``, Hessian of features from previous layer with respect to network input :math:`x` with shape :math:`(n_s, d, d, n_{in})`
        :type d2ud2x: torch.Tensor or ``None``
        :param kwargs: additional options, such as ``forward_mode`` as a user input
        :return:

            - **f** (*torch.Tensor*) - output features of network with shape :math:`(n_s, m)` where :math:`m` is the number of network output features
            - **dfdx** (*torch.Tensor* or ``None``) - if ``forward_mode = True``, gradient of output features with respect to network input :math:`x` with shape :math:`(n_s, d, m)`
            - **d2fd2x** (*torch.Tensor* or ``None``) - if ``forward_mode = True``, Hessian of output features with respect to network input :math:`x` with shape :math:`(n_s, d, d, m)`
        """
        forward_mode = self.setup_forward_mode(**kwargs)

        for module in self:
            x, dudx, d2ud2x = module(x, do_gradient=do_gradient, do_Hessian=do_Hessian, dudx=dudx, d2ud2x=d2ud2x,
                                     forward_mode=True if forward_mode is True else None)

        if (do_gradient or do_Hessian) and forward_mode is False:
            dudx, d2ud2x = self.backward(do_Hessian=do_Hessian)

        return x, dudx, d2ud2x

    def backward(self, do_Hessian: bool = False, dgdf: Union[torch.Tensor, None] = None,
                 d2gd2f: Union[torch.Tensor, None] = None) -> Tuple[torch.Tensor, Union[torch.Tensor, None]]:
        r"""
        Compute derivatives using backward propagation.  This method is called during the forward pass if ``forward_mode = False``.

        :param do_Hessian: If set to ``True``, the Hessian will be computed during the forward call. Default: ``False``
        :type do_Hessian: bool, optional
        :param dgdf: gradient of the subsequent layer features, :math:`g(f)`, with respect to the layer outputs, :math:`f` with shape :math:`(n_s, n_{out}, m)`.
        :type dgdf: torch.Tensor
        :param d2gd2f: gradient of the subsequent layer features, :math:`g(f)`, with respect to the layer outputs, :math:`f` with shape :math:`(n_s, n_{out}, n_{out}, m)`.
        :type d2gd2f: torch.Tensor or ``None``
        :return:

            - **dgdf** (*torch.Tensor* or ``None``) - gradient of the network with respect to input features :math:`x` with shape :math:`(n_s, d, m)`
            - **d2gd2f** (*torch.Tensor* or ``None``) - Hessian of the network with respect to input features :math:`u` with shape :math:`(n_s, d, d, m)`

        """
        for i in range(len(self) - 1, -1, -1):
            dgdf, d2gd2f = self[i].backward(do_Hessian=do_Hessian, dgdf=dgdf, d2gd2f=d2gd2f)
        return dgdf, d2gd2f


class NNPytorchAD(nn.Module):
    r"""
    Compute the derivatives of a network using Pytorch's automatic differentiation.

    The implementation follows that of `CP Flow`_.

    .. _CP Flow: https://github.com/CW-Huang/CP-Flow
    """

    def __init__(self, net: NN):
        r"""
        Create wrapper around hessQuik network.

        :param net: hessQuik network
        :type net: hessQuik.networks.NN
        """
        super(NNPytorchAD, self).__init__()
        self.net = net

    def forward(self, x: torch.Tensor, do_gradient: bool = False, do_Hessian: bool = False, **kwargs) \
            -> Tuple[torch.Tensor, Union[torch.Tensor, None], Union[torch.Tensor, None]]:
        r"""
        Forward propagate through the hessQuik network without computing derivatives.
        Then, use automatic differentiation to compute derivatives using ``torch.autograd.grad``.
        """

        (df, d2f) = (None, None)
        if do_gradient or do_Hessian:
            x.requires_grad = True

        # forwaard propagate without compute derivatives
        f, *_ = self.net(x, do_gradient=False, do_Hessian=False, forward_mode=False)

        if do_gradient or do_Hessian:
            f = f.view(x.shape[0], -1)
            df = []
            for j in range(f.shape[1]):
                df.append(grad(f[:, j].sum(), x, create_graph=True, retain_graph=True)[0])
            df = torch.stack(df, dim=2)

            if do_Hessian:
                df = df.reshape(x.shape[0], -1)
                d2f = []
                for j in range(df.shape[1]):
                    d2f.append(grad(df[:, j].sum(), x, create_graph=True, retain_graph=True)[0])
                d2f = torch.stack(d2f, dim=2)
                d2f = d2f.reshape(x.shape[0], x.shape[1], x.shape[1], -1).squeeze(-1)
                if d2f.dim() < 4:
                    d2f = d2f.unsqueeze(-1)

            df = df.reshape(x.shape[0], x.shape[1], -1).squeeze(-1)

            if df.dim() < 3:
                df = df.unsqueeze(-1)

        return f, df, d2f


class NNPytorchHessian(nn.Module):
    r"""
    Compute the derivatives of a network using Pytorch's Hessian functional.
    """

    def __init__(self, net):
        """
        Create wrapper around hessQuik network.

        :param net: hessQuik network
        :type net: hessQuik.networks.NN
        """
        super(NNPytorchHessian, self).__init__()
        self.net = net

    def forward(self, x: torch.Tensor, do_gradient: bool = False, do_Hessian: bool = False, **kwargs) \
            -> Tuple[torch.Tensor, Union[torch.Tensor, None], Union[torch.Tensor, None]]:
        r"""
        Forward propagate through the hessQuik network without computing derivatives.
        Then, use automatic differentiation to compute derivatives using ``torch.autograd.functional.hessian``.
        """

        (df, d2f) = (None, None)
        if do_gradient or do_Hessian:
            x.requires_grad = True

        f, *_ = self.net(x, do_gradient=False, do_Hessian=False, forward_mode=False)

        if f.squeeze().ndim > 1:
            raise ValueError(type(self), " must have scalar outputs per example")

        if do_gradient:
            df = grad(f.sum(), x)[0]
            df = df.unsqueeze(-1)

        if do_Hessian:
            d2f = hessian(lambda x: self.net(x)[0].sum(), x).sum(dim=2)
            d2f = d2f.unsqueeze(-1)

        return f, df, d2f


if __name__ == '__main__':
    import torch
    import hessQuik.activations as act
    import hessQuik.layers as lay
    from hessQuik.utils import input_derivative_check
    torch.set_default_dtype(torch.float64)

    # problem setup
    nex = 11
    d = 3
    ms = [2, 7, 5]
    m = 8
    x = torch.randn(nex, d)

    f = NN(lay.singleLayer(d, ms[0], act=act.softplusActivation()),
           lay.singleLayer(ms[0], ms[1], act=act.softplusActivation()),
           lay.singleLayer(ms[1], ms[2], act=act.softplusActivation()),
           lay.singleLayer(ms[2], m, act=act.softplusActivation()))

    # f = NNPytorchHessian(f)
    # x.requires_grad = True

    print('======= FORWARD =======')
    input_derivative_check(f, x, do_Hessian=True, verbose=True, forward_mode=True)

    print('======= BACKWARD =======')
    input_derivative_check(f, x, do_Hessian=True, verbose=True, forward_mode=False)



from setuptools import setup, find_packages
# from glob import glob

docs_extras = [
    'Sphinx >= 3.0.0',  # Force RTD to use >= 3.0.0
    'docutils',
    'pylons-sphinx-themes >= 1.0.8',  # Ethical Ads
    'pylons_sphinx_latesturl',
    'repoze.sphinx.autointerface',
    'sphinxcontrib-autoprogram',
]

setup(
    name='hessQuik',
    version='0.0.1',
    packages=find_packages(),
    url='https://github.com/elizabethnewman/hessQuik',
    license='MIT',
    author='Elizabeth Newman',
    author_email='elizabeth.newman@emory.edu',
    description='AD-free gradient and Hessian computations',
    install_requires=['torch'],
    extras_require={'interactive': ['numpy', 'matplotlib'], 'docs': docs_extras}
)



import torch
from math import log2, floor, ceil
import hessQuik
from hessQuik.utils import convert_to_base
from typing import Callable, Tuple, Optional, Union
import math

def input_derivative_check(f: Union[torch.nn.Module, Callable], x: torch.Tensor, do_Hessian: bool = False,
                           forward_mode: bool = True, num_test: int = 15, base: float = 2.0, tol: float = 0.1,
                           verbose: float = False) -> Tuple[Optional[bool], Optional[bool]]:
    r"""
    Taylor approximation test to verify derivatives.  Form the approximation by perturbing the input :math:`x`
    in the direction :math:`p` with step size :math:`h > 0` via

    .. math::

        f(x + h p) \approx f(x) + h\nabla f(x)^\top p + \frac{1}{2}p^\top \nabla^2f(x) p

    As :math:`h \downarrow 0^+`, the error between the approximation and the true value will decrease.
    The rate of decrease indicates the accuracy of the derivative computation.
    For details, see Chapter 5 of `Computational Methods for Electromagnetics`_ by Eldad Haber.

    .. _Computational Methods for Electromagnetics: https://epubs.siam.org/doi/book/10.1137/1.9781611973808

    Examples::

        >>> from hessQuik.layers import singleLayer
        >>> torch.set_default_dtype(torch.float64)  # use double precision to check implementations
        >>> x = torch.randn(10, 4)
        >>> f = singleLayer(4, 7, act=act.softplusActivation())
        >>> input_derivative_check(f, x, do_Hessian=True, verbose=True, forward_mode=True)
            h                           E0                      E1                      E2
            1.00 x 2^(00)		1.62 x 2^(-02)		1.70 x 2^(-07)		1.02 x 2^(-12)
            1.00 x 2^(-01)		1.63 x 2^(-03)		1.70 x 2^(-09)		1.06 x 2^(-15)
            1.00 x 2^(-02)		1.63 x 2^(-04)		1.69 x 2^(-11)		1.08 x 2^(-18)
            1.00 x 2^(-03)		1.63 x 2^(-05)		1.69 x 2^(-13)		1.09 x 2^(-21)
            1.00 x 2^(-04)		1.63 x 2^(-06)		1.69 x 2^(-15)		1.09 x 2^(-24)
            1.00 x 2^(-05)		1.63 x 2^(-07)		1.69 x 2^(-17)		1.10 x 2^(-27)
            1.00 x 2^(-06)		1.63 x 2^(-08)		1.69 x 2^(-19)		1.10 x 2^(-30)
            1.00 x 2^(-07)		1.63 x 2^(-09)		1.69 x 2^(-21)		1.10 x 2^(-33)
            1.00 x 2^(-08)		1.63 x 2^(-10)		1.69 x 2^(-23)		1.10 x 2^(-36)
            1.00 x 2^(-09)		1.63 x 2^(-11)		1.69 x 2^(-25)		1.10 x 2^(-39)
            1.00 x 2^(-10)		1.63 x 2^(-12)		1.69 x 2^(-27)		1.10 x 2^(-42)
            1.00 x 2^(-11)		1.63 x 2^(-13)		1.69 x 2^(-29)		1.10 x 2^(-45)
            1.00 x 2^(-12)		1.63 x 2^(-14)		1.69 x 2^(-31)		1.15 x 2^(-48)
            1.00 x 2^(-13)		1.63 x 2^(-15)		1.69 x 2^(-33)		1.33 x 2^(-50)
            1.00 x 2^(-14)		1.63 x 2^(-16)		1.69 x 2^(-35)		1.70 x 2^(-51)
            Gradient PASSED!
            Hessian PASSED!

    :param f: callable function that returns value, gradient, and Hessian
    :type f: torch.nn.Module or Callable
    :param x: input data
    :type x: torch.Tensor
    :param do_Hessian: If set to ``True``, the Hessian will be computed during the forward call. Default: ``False``
    :type do_Hessian: bool, optional
    :param forward_mode:  If set to ``False``, the derivatives will be computed in backward mode. Default: ``True``
    :type forward_mode: bool, optional
    :param num_test: number of perturbations
    :type num_test: int
    :param base: step size :math:`h = base^k`
    :type base: float
    :param tol: small tolerance to account for numerical errors when computing the order of approximation
    :type tol: float
    :param verbose: printout flag
    :type verbose: bool
    :return:
            - **grad_check** (*bool*) - if ``True``, gradient check passes
            - **hess_check** (*bool, optional*) -  if ``True``, Hessian check passes

    """
    # initial evaluation
    f0, df0, d2f0 = f(x, do_gradient=True, do_Hessian=do_Hessian, forward_mode=forward_mode)

    # ---------------------------------------------------------------------------------------------------------------- #
    # directional derivatives
    dx = torch.randn_like(x)
    dx = dx / torch.norm(x)
    curvx = None
    if isinstance(f, hessQuik.activations.hessQuikActivationFunction):
        dfdx = df0 * dx

        if d2f0 is not None:
            curvx = torch.sum(dx.unsqueeze(0) * d2f0 * dx.unsqueeze(0), dim=0)
    else:
        dfdx = torch.matmul(df0.transpose(1, 2), dx.unsqueeze(2)).squeeze(2)

        if d2f0 is not None:
            curvx = torch.sum(dx.unsqueeze(2).unsqueeze(3) * d2f0 * dx.unsqueeze(1).unsqueeze(3), dim=(1, 2))

    # ---------------------------------------------------------------------------------------------------------------- #
    # derivative check
    grad_check, hess_check = None, None
    E0, E1, E2 = [], [], []

    if verbose:
        headers = ('h', 'E0', 'E1')
        if do_Hessian:
            headers += ('E2',)
        print(('{:<20s}' * len(headers)).format(*headers))

    for k in range(num_test):
        h = base ** (-k)
        ft, *_ = f(x + h * dx, do_gradient=False, do_Hessian=False)

        E0.append(torch.norm(f0 - ft).item())
        E1.append(torch.norm(f0 + h * dfdx - ft).item())
        printouts = convert_to_base((E0[-1], E1[-1]))

        if curvx is not None:
            E2.append(torch.norm(f0 + h * dfdx + 0.5 * (h ** 2) * curvx - ft).item())
            printouts += convert_to_base((E2[-1],))

        if verbose:
            print(((1 + len(printouts) // 2) * '%0.2f x 2^(%0.2d)\t\t') % ((1, -k) + printouts))

    E0, E1, E2 = torch.tensor(E0), torch.tensor(E1), torch.tensor(E2)

    # ---------------------------------------------------------------------------------------------------------------- #
    # check if order is 2 enough of the time
    eps = torch.finfo(x.dtype).eps
    grad_check = (sum((torch.log2(E1[:-1] / E1[1:]) / log2(base)) > (2 - tol)) > num_test // 3)
    grad_check = (grad_check or (torch.kthvalue(E1, num_test // 4)[0] < (100 * eps)))

    if curvx is not None:
        hess_check = (sum((torch.log2(E2[:-1] / E2[1:]) / log2(base)) > (3 - tol)) > num_test // 3)
        hess_check = (hess_check or (torch.kthvalue(E2, num_test // 4)[0] < (100 * eps)))

    if verbose:
        if grad_check:
            print('Gradient PASSED!')
        else:
            print('Gradient FAILED.')

        if curvx is not None:
            if hess_check:
                print('Hessian PASSED!')
            else:
                print('Hessian FAILED.')

    return grad_check, hess_check


def input_derivative_check_finite_difference(f: Callable, x: torch.Tensor,
                                             do_Hessian: bool = False, forward_mode: bool = True,
                                             eps: float = 1e-4, atol: float = 1e-5, rtol: float = 1e-3,
                                             verbose: bool = False) -> Tuple[Optional[bool], Optional[bool]]:
    r"""
    Finite difference test to verify derivatives.  Form the approximation by perturbing each entry in the input
    in the unit direction with step size :math:`\varepsilon > 0`:

    .. math::

        \widetilde{\nabla f}_{i} = \frac{f(x_i + \varepsilon) -  f(x_i - \varepsilon)}{2\varepsilon}

    where :math:`x_i \pm \varepsilon` means add or subtract :math:`\varepsilon` from
    the i-th entry of the input :math:`x`, but leave the other entries unchanged.
    The notation :math:`\widetilde{(\cdot)}` indicates the finite difference approximation.

    Examples::

        >>> from hessQuik.layers import singleLayer
        >>> torch.set_default_dtype(torch.float64)  # use double precision to check implementations
        >>> x = torch.randn(10, 4)
        >>> f = singleLayer(4, 7, act=act.tanhActivation())
        >>> input_derivative_check_finite_difference(f, x, do_Hessian=True, verbose=True, forward_mode=True)
            Gradient Finite Difference: Error = 8.1720e-10, Relative Error = 2.5602e-10
            Gradient PASSED!
            Hessian Finite Difference: Error = 4.5324e-08, Relative Error = 4.4598e-08
            Hessian PASSED!

    :param f: callable function that returns value, gradient, and Hessian
    :type f: Callable
    :param x: input data
    :type x: torch.Tensor
    :param do_Hessian: If set to ``True``, the Hessian will be computed during the forward call. Default: ``False``
    :type do_Hessian: bool, optional
    :param forward_mode:  If set to ``False``, the derivatives will be computed in backward mode. Default: ``True``
    :type forward_mode: bool, optional
    :param eps: step size. Default: 1e-4
    :type eps: float
    :param atol: absolute tolerance, e.g., :math:`\|\nabla f - \widetilde{\nabla f}\| < atol`. Default: 1e-5
    :type atol: float
    :param rtol: relative tolerance, e.g., :math:`\|\nabla f - \widetilde{\nabla f}\|/\|\nabla f\| < rtol`. Default: 1e-3
    :type rtol: float
    :param verbose: printout flag
    :type verbose: bool
    :return:
            - **grad_check** (*bool*) - if ``True``, gradient check passes
            - **hess_check** (*bool, optional*) -  if ``True``, Hessian check passes
    """

    # compute initial gradient
    f0, df0, d2f0 = f(x, do_gradient=True, do_Hessian=do_Hessian, forward_mode=forward_mode)

    d = x.shape[1]

    # ---------------------------------------------------------------------------------------------------------------- #
    # test gradient
    df0_approx = torch.zeros_like(df0)
    for i in range(d):
        # perturbation in standard directions
        ei = torch.zeros(d, device=f0.device, dtype=f0.dtype)
        ei[i] = eps

        f_pos, *_ = f(x + ei.unsqueeze(0))
        f_neg, *_ = f(x - ei.unsqueeze(0))

        df0_approx[:, i] = (f_pos - f_neg) / (2 * eps)

    err = torch.norm(df0 - df0_approx).item()
    rel_err = err / torch.norm(df0).item()

    grad_check = (err < atol and rel_err < rtol)
    if verbose:
        print('Gradient Finite Difference: Error = %0.4e, Relative Error = %0.4e' % (err, rel_err))

        if grad_check:
            print('Gradient PASSED!')
        else:
            print('Gradient FAILED.')

    # ---------------------------------------------------------------------------------------------------------------- #
    # test Hessian
    # https://v8doc.sas.com/sashtml/ormp/chap5/sect28.htm
    hess_check = None
    if do_Hessian:

        d2f0_approx = torch.zeros_like(d2f0)
        for i in range(d):
            ei = torch.zeros(d, device=f0.device, dtype=f0.dtype)
            ei[i] = eps

            for j in range(d):
                ej = torch.zeros(d, device=f0.device, dtype=f0.dtype)
                ej[j] = eps

                f1, *_ = f(x + (ei + ej).unsqueeze(0))
                f2, *_ = f(x + (ei - ej).unsqueeze(0))
                f3, *_ = f(x + (-ei + ej).unsqueeze(0))
                f4, *_ = f(x - (ei + ej).unsqueeze(0))

                d2f0_approx[:, i, j] = (f1 - f2 - f3 + f4) / (4 * (eps ** 2))

        err = torch.norm(d2f0 - d2f0_approx).item()
        rel_err = err / max(torch.norm(d2f0).item(), eps)

        hess_check = (err < atol and rel_err < rtol)
        if verbose:
            print('Hessian Finite Difference: Error = %0.4e, Relative Error = %0.4e' % (err, rel_err))

            if hess_check:
                print('Hessian PASSED!')
            else:
                print('Hessian FAILED.')

    return grad_check, hess_check


if __name__ == '__main__':
    from hessQuik.networks import NN
    from hessQuik.layers import singleLayer
    import hessQuik.activations as act
    torch.set_default_dtype(torch.float64)

    nex = 11  # no. of examples
    d = 4  # no. of input features

    x = torch.randn(nex, d)
    dx = torch.randn_like(x)

    f = NN(singleLayer(d, 7, act=act.identityActivation()),
           singleLayer(7, 5, act=act.identityActivation()))

    input_derivative_check(f, x, do_Hessian=True, verbose=True)

    input_derivative_check_finite_difference(f, x, do_Hessian=True, verbose=True, forward_mode=False)




import torch
from math import log2, floor
import hessQuik
from hessQuik.utils import convert_to_base, extract_data, insert_data
from typing import Callable, Tuple, Optional, Union


def network_derivative_check(f: torch.nn.Module, x: torch.Tensor, do_Hessian: bool = False,
                             forward_mode: bool = True, num_test: int = 15, base: float = 2.0, tol: float = 0.1,
                             verbose: bool = False) -> Optional[bool]:
    r"""
    Taylor approximation test to verify derivatives.  Form the approximation by perturbing the
    network weights :math:`\theta` in the direction :math:`p` with step size :math:`h > 0` via

    .. math::

        \Phi(\theta + h p) \approx \Phi(\theta) + h\nabla_{\theta} \Phi(\theta)^\top p

    where :math:`\Phi` is the objective function and :math:`\theta` are the network weights.  This test uses the loss

    .. math::

        \Phi(\theta) = \frac{1}{2}\|f_{\theta}(x)\|^2 + \frac{1}{2}\|\nabla f_{\theta}(x)\|^2 +  \frac{1}{2}\|\nabla^2 f_{\theta}(x)\|^2

    to validate network gradient computation after computing derivatives of the
    input features of the network :math:`f_{\theta}`.

    As :math:`h \downarrow 0^+`, the error between the approximation and the true value will decrease.
    The rate of decrease indicates the accuracy of the derivative computation.
    For details, see Chapter 5 of `Computational Methods for Electromagnetics`_ by Eldad Haber.

    .. _Computational Methods for Electromagnetics: https://epubs.siam.org/doi/book/10.1137/1.9781611973808

    Examples::

        >>> import hessQuik.activations as act, hessQuik.layers as lay, hessQuik.networks as net
        >>> torch.set_default_dtype(torch.float64)  # use double precision to check implementations
        >>> x = torch.randn(10, 4)
        >>> width, depth = 8, 3
        >>> f = net.NN(lay.singleLayer(4, width, act=act.tanhActivation()), net.resnetNN(width, depth, h=1.0, act=act.tanhActivation()), lay.singleLayer(width, 1, act=act.identityActivation()))
        >>> network_derivative_check(f, x, do_Hessian=True, verbose=True, forward_mode=True)
            h                           E0                      E1
            1.00 x 2^(00)		1.97 x 2^(02)		1.05 x 2^(01)
            1.00 x 2^(-01)		1.14 x 2^(02)		1.71 x 2^(-02)
            1.00 x 2^(-02)		1.20 x 2^(01)		1.50 x 2^(-04)
            1.00 x 2^(-03)		1.22 x 2^(00)		1.40 x 2^(-06)
            1.00 x 2^(-04)		1.24 x 2^(-01)		1.35 x 2^(-08)
            1.00 x 2^(-05)		1.24 x 2^(-02)		1.32 x 2^(-10)
            1.00 x 2^(-06)		1.24 x 2^(-03)		1.31 x 2^(-12)
            1.00 x 2^(-07)		1.24 x 2^(-04)		1.30 x 2^(-14)
            1.00 x 2^(-08)		1.25 x 2^(-05)		1.30 x 2^(-16)
            1.00 x 2^(-09)		1.25 x 2^(-06)		1.30 x 2^(-18)
            1.00 x 2^(-10)		1.25 x 2^(-07)		1.30 x 2^(-20)
            1.00 x 2^(-11)		1.25 x 2^(-08)		1.30 x 2^(-22)
            1.00 x 2^(-12)		1.25 x 2^(-09)		1.30 x 2^(-24)
            1.00 x 2^(-13)		1.25 x 2^(-10)		1.30 x 2^(-26)
            1.00 x 2^(-14)		1.25 x 2^(-11)		1.30 x 2^(-28)
            Gradient PASSED!

    :param f: callable function that returns value, gradient, and Hessian
    :type f: torch.nn.Module
    :param x: input data
    :type x: torch.Tensor
    :param do_Hessian: If set to ``True``, the Hessian will be computed during the forward call. Default: ``False``
    :type do_Hessian: bool, optional
    :param forward_mode:  If set to ``False``, the derivatives will be computed in backward mode. Default: ``True``
    :type forward_mode: bool, optional
    :param num_test: number of perturbations
    :type num_test: int
    :param base: step size :math:`h = base^k`
    :type base: float
    :param tol: small tolerance to account for numerical errors when computing the order of approximation
    :type tol: float
    :param verbose: printout flag
    :type verbose: bool
    :return:
            - **grad_check** (*bool*) - if ``True``, gradient check passes

    """
    loss_df, loss_d2f = 0.0, 0.0

    # initial evaluation
    f0, df0, d2f0 = f(x, do_gradient=True, do_Hessian=do_Hessian, forward_mode=forward_mode)

    # compute loss
    loss_f = 0.5 * torch.norm(f0) ** 2

    if df0 is not None:
        loss_df = 0.5 * torch.norm(df0) ** 2

    if d2f0 is not None:
        loss_d2f = 0.5 * torch.norm(d2f0) ** 2

    loss = loss_f + loss_df + loss_d2f
    loss.backward()

    loss0 = loss.detach()
    theta0 = extract_data(f, 'data')
    grad_theta0 = extract_data(f, 'grad')

    # perturbation
    dtheta = torch.randn_like(theta0)
    dtheta = dtheta / torch.norm(dtheta)

    # directional derivative
    dfdtheta = (grad_theta0 * dtheta).sum()

    # ---------------------------------------------------------------------------------------------------------------- #
    # derivative check
    if verbose:
        headers = ('h', 'E0', 'E1')
        print(('{:<20s}' * len(headers)).format(*headers))

    with torch.no_grad():
        E0, E1 = [], []
        loss_dft, loss_d2ft = 0.0, 0.0
        for k in range(num_test):
            h = base ** (-k)
            insert_data(f, theta0 + h * dtheta)
            ft, dft, d2ft = f(x, do_gradient=True, do_Hessian=do_Hessian)

            # compute loss
            loss_ft = 0.5 * torch.norm(ft) ** 2

            if df0 is not None:
                loss_dft = 0.5 * torch.norm(dft) ** 2

            if d2f0 is not None:
                loss_d2ft = 0.5 * torch.norm(d2ft) ** 2

            losst = loss_ft + loss_dft + loss_d2ft
            E0.append(torch.norm(loss0 - losst).item())
            E1.append(torch.norm(loss0 + h * dfdtheta - losst).item())

            printouts = convert_to_base((E0[-1], E1[-1]))

            if verbose:
                print(((1 + len(printouts) // 2) * '%0.2f x 2^(%0.2d)\t\t') % ((1, -k) + printouts))

    E0, E1 = torch.tensor(E0), torch.tensor(E1)

    # ---------------------------------------------------------------------------------------------------------------- #
    # check if order is 2 at least half of the time
    eps = torch.finfo(x.dtype).eps
    grad_check = (sum((torch.log2(E1[:-1] / E1[1:]) / log2(base)) > (2 - tol)) > 3)
    grad_check = (grad_check or (torch.kthvalue(E1, num_test // 3)[0] < (100 * eps)))

    if verbose:
        if grad_check:
            print('Gradient PASSED!')
        else:
            print('Gradient FAILED.')

    return grad_check


if __name__ == '__main__':
    import hessQuik.networks as net
    import hessQuik.layers as lay
    import hessQuik.activations as act
    torch.set_default_dtype(torch.float64)

    nex = 11  # no. of examples
    d = 2  # no. of input features

    x = torch.randn(nex, d)
    dx = torch.randn_like(x)

    # f = net.NN(lay.singleLayer(d, 7, act=act.softplusActivation()),
    #            lay.singleLayer(7, 5, act=act.identityActivation()))

    width = 8
    depth = 8
    f = net.NN(lay.singleLayer(d, width, act=act.tanhActivation()),
               net.resnetNN(width, depth, h=1.0, act=act.tanhActivation()),
               lay.singleLayer(width, 1, act=act.identityActivation()))

    # width = 7
    # f = net.NN(lay.singleLayer(d, width, act=act.tanhActivation()),
    #            net.resnetNN(width, 4, act=act.softplusActivation()),
    #            net.fullyConnectedNN([width, 13, 5], act=act.quadraticActivation()),
    #            lay.singleLayer(5, 3, act=act.identityActivation()),
    #            lay.quadraticLayer(3, 2)
    #            )

    network_derivative_check(f, x, do_Hessian=True, verbose=True)




import torch.nn as nn
from torch import Tensor
from typing import Union, Tuple


class hessQuikLayer(nn.Module):
    r"""
    Base class for all hessQuik layers.
    """

    def __init__(self, *args, **kwargs):
        super(hessQuikLayer, self).__init__()

    def dim_input(self) -> int:
        r"""

        :return: dimension of input features
        :rtype: int
        """
        raise NotImplementedError

    def dim_output(self) -> int:
        r"""

        :return: dimension of output features
        :rtype: int
        """
        raise NotImplementedError

    def forward(self, u: Tensor, do_gradient: bool = False, do_Hessian: bool = False, forward_mode: bool = True,
                dudx: Union[Tensor, None] = None, d2ud2x: Union[Tensor, None] = None) \
            -> Tuple[Tensor, Union[Tensor, None], Union[Tensor, None]]:
        r"""
        Forward pass through the layer that maps input features :math:`u` of size :math:`(n_s, n_{in})`
        to output features :math:`f` of size :math:`(n_s, n_{out})` where :math:`n_s` is the number of samples,
        :math:`n_{in}` is the number of input features, and :math:`n_{out}` is the number of output features.

        The input features :math:`u(x)` is a function of the network input :math:`x` of size :math:`(n_s, d)`
        where :math:`d` is the dimension of the network input.

        :param u: features from previous layer with shape :math:`(n_s, n_{in})`
        :type u: torch.Tensor
        :param do_gradient: If set to ``True``, the gradient will be computed during the forward call. Default: ``False``
        :type do_gradient: bool, optional
        :param do_Hessian: If set to ``True``, the Hessian will be computed during the forward call. Default: ``False``
        :type do_Hessian: bool, optional
        :param forward_mode:  If set to ``False``, the derivatives will be computed in backward mode. Default: ``True``
        :type forward_mode: bool, optional
        :param dudx: if ``forward_mode = True``, gradient of features from previous layer with respect to network input :math:`x` with shape :math:`(n_s, d, n_{in})`
        :type dudx: torch.Tensor or ``None``
        :param d2ud2x: if ``forward_mode = True``, Hessian of features from previous layer with respect to network input :math:`x` with shape :math:`(n_s, d, d, n_{in})`
        :type d2ud2x: torch.Tensor or ``None``
        :return:

            - **f** (*torch.Tensor*) - output features of layer with shape :math:`(n_s, n_{out})`
            - **dfdx** (*torch.Tensor* or ``None``) - if ``forward_mode = True``, gradient of output features with respect to network input :math:`x` with shape :math:`(n_s, d, n_{out})`
            - **d2fd2x** (*torch.Tensor* or ``None``) - if ``forward_mode = True``, Hessian of output features with respect to network input :math:`x` with shape :math:`(n_s, d, d, n_{out})`

        """

        raise NotImplementedError

    def backward(self, do_Hessian: bool = False,
                 dgdf: Union[Tensor, None] = None, d2gd2f: Union[Tensor, None] = None) \
            -> Tuple[Tensor, Union[Tensor, None]]:
        r"""
        Backward pass through the layer that maps the gradient of the network :math:`g`
        with respect to the output features :math:`f` of size :math:`(n_s, n_{out}, m)`
        to the gradient of the network :math:`g` with respect to the input features :math:`u` of size :math:`(n_s, n_{in}, m)`
        where :math:`n_s` is the number of samples, :math:`n_{in}` is the number of input features,,
        :math:`n_{out}` is the number of output features, and :math:`m` is the number of network output features.

        :param do_Hessian: If set to ``True``, the Hessian will be computed during the forward call. Default: ``False``
        :type do_Hessian: bool, optional
        :param dgdf: gradient of the subsequent layer features, :math:`g(f)`, with respect to the layer outputs, :math:`f` with shape :math:`(n_s, n_{out}, m)`.
        :type dgdf: torch.Tensor
        :param d2gd2f: gradient of the subsequent layer features, :math:`g(f)`, with respect to the layer outputs, :math:`f` with shape :math:`(n_s, n_{out}, n_{out}, m)`.
        :type d2gd2f: torch.Tensor or ``None``
        :return:

            - **dgdu** (*torch.Tensor* or ``None``) -  gradient of the network with respect to input features :math:`u` with shape :math:`(n_s, n_{in}, m)`
            - **d2gd2u** (*torch.Tensor* or ``None``) - Hessian of the network with respect to input features :math:`u` with shape :math:`(n_s, n_{in}, n_{in}, m)`

        """

        raise NotImplementedError





import torch
from hessQuik.layers import hessQuikLayer
import hessQuik.activations as act
from hessQuik.layers import singleLayer
from typing import Union, Tuple


class resnetLayer(hessQuikLayer):
    r"""
    Evaluate and compute derivatives of a residual layer.

    Examples::

        >>> import hessQuik.layers as lay
        >>> f = lay.resnetLayer(4, h=0.25)
        >>> x = torch.randn(10, 4)
        >>> fx, dfdx, d2fd2x = f(x, do_gradient=True, do_Hessian=True)
        >>> print(fx.shape, dfdx.shape, d2fd2x.shape)
        torch.Size([10, 4]) torch.Size([10, 4, 4]) torch.Size([10, 4, 4, 4])

    """

    def __init__(self, width: int, h: float = 1.0, act: act.hessQuikActivationFunction = act.identityActivation(),
                 device=None, dtype=None) -> None:
        r"""
        :param width: number of input and output features, :math:`w`
        :type width: int
        :param h: step size, :math:`h > 0`
        :type h: float
        :param act: activation function
        :type act: hessQuikActivationFunction

        :var layer: singleLayer with :math:`w` input features and :math:`w` output features
        """
        factory_kwargs = {'device': device, 'dtype': dtype}
        super(resnetLayer, self).__init__()

        self.width = width
        self.h = h
        self.layer = singleLayer(width, width, act=act, **factory_kwargs)

    def dim_input(self) -> int:
        r"""
        width
        """
        return self.width

    def dim_output(self) -> int:
        r"""
        width
        """
        return self.width

    def forward(self, u: torch.Tensor, do_gradient: bool = False, do_Hessian: bool = False, forward_mode: bool = True,
                dudx: Union[torch.Tensor, None] = None, d2ud2x: Union[torch.Tensor, None] = None) \
            -> Tuple[torch.Tensor, Union[torch.Tensor, None], Union[torch.Tensor, None]]:
        r"""
        Forward propagation through resnet layer of the form

        .. math::

            f(x) = u(x) + h \cdot singleLayer(u(x))

        Here, :math:`u(x)` is the input into the layer of size :math:`(n_s, w)` which is
        a function of the input of the network, :math:`x`.
        The output features, :math:`f(x)`, are of size :math:`(n_s, w)`.

        As an example, for one sample, :math:`n_s = 1`, the gradient with respect to :math:`x` is of the form

        .. math::

            \nabla_x f = I + h \nabla_x singleLayer(u(x))

        where :math:`I` denotes the :math:`w \times w` identity matrix.
        """
        (dfdx, d2fd2x) = (None, None)
        fi, dfi, d2fi = self.layer(u, do_gradient=do_gradient, do_Hessian=do_Hessian, dudx=dudx, d2ud2x=d2ud2x,
                                   forward_mode=True if forward_mode is True else None)

        # skip connection
        f = u + self.h * fi

        if do_gradient and forward_mode is True:

            if dudx is None:
                dfdx = torch.eye(self.width, dtype=dfi.dtype, device=dfi.device) + self.h * dfi
            else:
                dfdx = dudx + self.h * dfi

        if do_Hessian and forward_mode is True:
            d2fd2x = self.h * d2fi
            if d2ud2x is not None:
                d2fd2x += d2ud2x

        if (do_gradient or do_Hessian) and forward_mode is False:
            dfdx, d2fd2x = self.backward(do_Hessian=do_Hessian)

        return f, dfdx, d2fd2x

    def backward(self, do_Hessian: bool = False,
                 dgdf: Union[torch.Tensor, None] = None, d2gd2f: Union[torch.Tensor, None] = None):
        r"""
        Backward propagation through single layer of the form

        .. math::

                f(u) = u + h \cdot singleLayer(u)

        Here, the network is :math:`g` is a function of :math:`f(u)`.

        As an example, for one sample, :math:`n_s = 1`, the gradient of the network with respect to :math:`u` is of the form

        .. math::

            \nabla_u g = \nabla_f g + h \cdot \nabla_u singleLayer(u)

        where :math:`\odot` denotes the pointwise product.

        """
        
        d2gd2u = None
        if not do_Hessian:

            dgdu = self.layer.backward(do_Hessian=False, dgdf=dgdf, d2gd2f=None)[0]

            if dgdf is None:
                dgdu = torch.eye(self.width, dtype=dgdu.dtype, device=dgdu.device) + self.h * dgdu
            else:
                dgdu = dgdf + self.h * dgdu
        else:
            dfdx, d2fd2x = self.layer.backward(do_Hessian=do_Hessian, dgdf=None, d2gd2f=None)[:2]

            dgdu = torch.eye(self.width, dtype=dfdx.dtype, device=dfdx.device) + self.h * dfdx
            if dgdf is not None:
                dgdu = dgdu @ dgdf

            # d2gd2u = self.h * d2fd2x
            if d2gd2f is None:
                d2gd2u = self.h * d2fd2x
            else:
                # TODO: compare timings for h_dfdx on CPU and GPU
                h_dfdx = torch.eye(self.width, dtype=dfdx.dtype, device=dfdx.device) + self.h * dfdx

                # Gauss-Newton approximation
                h1 = (h_dfdx.unsqueeze(1) @ d2gd2f.permute(0, 3, 1, 2) @ h_dfdx.permute(0, 2, 1).unsqueeze(1))
                h1 = h1.permute(0, 2, 3, 1)

                # extra term to compute full Hessian
                h2 = d2fd2x @ dgdf.unsqueeze(1)
                # combine
                d2gd2u = h1 + self.h * h2

        return dgdu, d2gd2u

    def extra_repr(self) -> str:
        r"""
        :meta private:
        """
        return 'width={}, h={}'.format(self.width, self.h)


if __name__ == '__main__':
    from hessQuik.utils import input_derivative_check
    torch.set_default_dtype(torch.float64)

    nex = 11  # no. of examples
    width = 4  # no. of input features
    h = 0.25
    x = torch.randn(nex, width)
    f = resnetLayer(width, h=h, act=act.softplusActivation())

    f0, df0, d2f0 = f(x, do_gradient=True, do_Hessian=True, forward_mode=False)

    print('======= FORWARD =======')
    input_derivative_check(f, x, do_Hessian=True, verbose=True, forward_mode=True)

    print('======= BACKWARD =======')
    input_derivative_check(f, x, do_Hessian=True, verbose=True, forward_mode=False)